from __future__ import annotations

from typing import Any

from x8.core import DataModel, Response, operation
from x8.ql import Expression, OrderBy, Select, Update
from x8.storage._common import (
    CollectionResult,
    Index,
    IndexResult,
    StoreComponent,
)

from ._models import (
    SearchBatch,
    SearchCollectionConfig,
    SearchItem,
    SearchKey,
    SearchList,
)


class SearchStore(StoreComponent):
    collection: str | None

    def __init__(
        self,
        collection: str | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            collection:
                Default collection name.
        """
        self.collection = collection
        super().__init__(**kwargs)

    @operation()
    def create_collection(
        self,
        collection: str | None = None,
        config: dict | SearchCollectionConfig | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.
            where:
                Condition expression.

        Returns:
            Collection result.
        """
        raise NotImplementedError

    @operation()
    def drop_collection(
        self,
        collection: str | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Drop collection.

        Args:
            collection:
                Collection name.
            where:
                Condition expression.

        Returns:
            Collection result.
        """
        raise NotImplementedError

    @operation()
    def list_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        raise NotImplementedError

    @operation()
    def has_collection(
        self,
        collection: str | None = None,
    ) -> Response[bool]:
        """Check if collection exists.

        Args:
            collection:
                Collection name.

        Returns:
            A value indicating whether collection exists.
        """
        raise NotImplementedError

    @operation()
    def create_index(
        self,
        index: dict | Index,
        collection: str | None = None,
    ) -> Response[IndexResult]:
        """Create index.

        Args:
            index:
                Index to create.
            collection:
                Collection name.

        Returns:
            Index result.
        """
        raise NotImplementedError

    @operation()
    def drop_index(
        self,
        index: dict | Index | None = None,
        collection: str | None = None,
    ) -> Response[IndexResult]:
        """Drop index.

        Args:
            index:
                Index to drop.
            collection:
                Collection name.

        Returns:
            Index result.
        """
        raise NotImplementedError

    @operation()
    def list_indexes(
        self,
        collection: str | None = None,
    ) -> Response[list[Index]]:
        """List indexes.

        Args:
            collection:
                Collection name.

        Returns:
            List of indexes.
        """
        raise NotImplementedError

    @operation()
    def get(
        self,
        key: str | dict | SearchKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[SearchItem]:
        """Get document.

        Args:
            key:
                Document key.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            NotFoundError:
                Item not found.
        """
        raise NotImplementedError

    @operation()
    def put(
        self,
        value: dict[str, Any] | DataModel,
        key: str | dict | SearchKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[SearchItem]:
        """Put document.

        Args:
            value:
                Document value.
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    def update(
        self,
        key: str | dict | SearchKey,
        set: str | Update,
        where: str | Expression | None = None,
        returning: str | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[SearchItem]:
        """Update document.

        Args:
            key:
                Document key.
            set:
                Update expression.
            where:
                Condition expression.
            returning:
                A value indicating whether updated document is returned.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            NotFoundError:
                Document not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    def delete(
        self,
        key: str | dict | SearchKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete document.

        Args:
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Document not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    def query(
        self,
        search: str | Expression | None = None,
        select: str | Select | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[SearchList]:
        """Query documents.

        Args:
            search:
                Search expression.
            select:
                Select expression.
            where:
                Condition expression.
            order_by:
                Order by expression.
            limit:
                Query limit.
            offset:
                Query offset.
            collection:
                Collection name.

        Returns:
            Document list with items.
        """
        raise NotImplementedError

    @operation()
    def count(
        self,
        search: str | Expression | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count documents.

        Args:
            search:
                Search expression.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Count of documents.
        """
        raise NotImplementedError

    @operation()
    def batch(
        self,
        batch: dict | SearchBatch,
        **kwargs,
    ) -> Response[list[Any]]:
        """Execute batch.

        Args:
            batch:
                Batch operation.

        Returns:
            Batch operation results.
        """
        raise NotImplementedError

    @operation()
    def close(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close client.

        Returns:
            None.
        """
        raise NotImplementedError

    @operation()
    async def acreate_collection(
        self,
        collection: str | None = None,
        config: dict | SearchCollectionConfig | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.
            where:
                Condition expression.

        Returns:
            Collection result.
        """
        raise NotImplementedError

    @operation()
    async def adrop_collection(
        self,
        collection: str | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Drop collection.

        Args:
            collection:
                Collection name.
            where:
                Condition expression.

        Returns:
            Collection result.
        """
        raise NotImplementedError

    @operation()
    async def alist_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        raise NotImplementedError

    @operation()
    async def ahas_collection(
        self, collection: str | None = None
    ) -> Response[bool]:
        """Check if collection exists.

        Args:
            collection:
                Collection name.

        Returns:
            A value indicating whether collection exists.
        """
        raise NotImplementedError

    @operation()
    async def acreate_index(
        self,
        index: dict | Index,
        collection: str | None = None,
    ) -> Response[IndexResult]:
        """Create index.

        Args:
            index:
                Index to create.
            collection:
                Collection name.

        Returns:
            Index result.
        """
        raise NotImplementedError

    @operation()
    async def adrop_index(
        self,
        index: dict | Index | None = None,
        collection: str | None = None,
    ) -> Response[IndexResult]:
        """Drop index.

        Args:
            index:
                Index to drop.
            collection:
                Collection name.

        Returns:
            Index result.
        """
        raise NotImplementedError

    @operation()
    async def alist_indexes(
        self,
        collection: str | None = None,
    ) -> Response[list[Index]]:
        """List indexes.

        Args:
            collection:
                Collection name.
        """
        raise NotImplementedError

    @operation()
    async def aget(
        self,
        key: str | dict | SearchKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[SearchItem]:
        """Get document.

        Args:
            key:
                Document key.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            NotFoundError:
                Document not found.
        """
        raise NotImplementedError

    @operation()
    async def aput(
        self,
        value: dict[str, Any] | DataModel,
        key: str | dict | SearchKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[SearchItem]:
        """Put document.

        Args:
            value:
                Document.
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    async def aupdate(
        self,
        key: str | dict | SearchKey,
        set: str | Update,
        where: str | Expression | None = None,
        returning: str | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[SearchItem]:
        """Update document.

        Args:
            key:
                Document key.
            set:
                Update expression.
            where:
                Condition expression.
            returning:
                A value indicating whether updated document is returned.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            NotFoundError:
                Document not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    async def adelete(
        self,
        key: str | dict | SearchKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete document.

        Args:
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Document not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    async def aquery(
        self,
        search: str | Expression | None = None,
        select: str | Select | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[SearchList]:
        """Query documents.

        Args:
            search:
                Search expression.
            select:
                Select expression.
            where:
                Condition expression.
            order_by:
                Order by expression.
            limit:
                Query limit.
            offset:
                Query offset.
            collection:
                Collection name.

        Returns:
            Document list with items.
        """
        raise NotImplementedError

    @operation()
    async def acount(
        self,
        search: str | Expression | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count documents.

        Args:
            search:
                Search expression.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Count of documents.
        """
        raise NotImplementedError

    @operation()
    async def abatch(
        self,
        batch: dict | SearchBatch,
        **kwargs: Any,
    ) -> Response[list[Any]]:
        """Execute batch.

        Args:
            batch:
                Batch operation.

        Returns:
            Batch operation results.
        """
        raise NotImplementedError

    @operation()
    async def aclose(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close async client.

        Returns:
            None.
        """
        raise NotImplementedError



from ._config import DEFAULT_BASE_IMAGE, DEFAULT_PLATFORM
from ._models import (
    BuildConfig,
    ContainerItem,
    ImageItem,
    PrepareConfig,
    RunConfig,
    SourceItem,
)
from .component import Containerizer

__all__ = [
    "Containerizer",
    "BuildConfig",
    "PrepareConfig",
    "RunConfig",
    "SourceItem",
    "ImageItem",
    "ContainerItem",
    "DEFAULT_BASE_IMAGE",
    "DEFAULT_PLATFORM",
]



from __future__ import annotations

import base64
import json
from typing import Any

from azure.core.exceptions import ResourceExistsError, ResourceNotFoundError
from azure.storage.queue import QueueMessage, QueueProperties

from x8._common.azure_provider import AzureProvider
from x8.core import Context, DataModel, NCall, Operation, Response
from x8.core.exceptions import BadRequestError, ConflictError, NotFoundError
from x8.messaging._common import (
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    MessagingOperation,
    MessagingOperationParser,
    QueueConfig,
    QueueInfo,
)

from .._feature import QueueFeature


class AzureQueueStorage(AzureProvider):
    account_name: str | None
    queue: str | None
    connection_string: str | None
    nparams: dict[str, Any]

    _client: Any
    _aclient: Any
    _queue_cache: dict[str, AzureQueue]
    _aqueue_cache: dict[str, AzureQueue]
    _op_converter: OperationConverter
    _result_converter: ResultConverter

    def __init__(
        self,
        queue: str | None = None,
        connection_string: str | None = None,
        account_url: str | None = None,
        account_key: str | None = None,
        sas_token: str | None = None,
        credential_type: str | None = "default",
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs: Any,
    ):
        """Initialize.

        Args:
            queue:
                Queue name.
            connection_string:
                Azure storage connection string.
            account_url:
                Azure storage account url.
            account_key:
                Azure storage account key.
            sas_token:
                SAS token.
            credential_type:
                Azure credential type.
            tenant_id:
                Azure tenant id for client_secret credential type.
            client_id:
                Azure client id for client_secret credential type.
            client_secret:
                Azure client secret for client_secret credential type.
            certificate_path:
                Certificate path for certificate credential type.
            nparams:
                Native parameters to Queue Storage client.
        """
        self.queue = queue
        self.connection_string = connection_string
        self.account_url = account_url
        self.account_key = account_key
        self.sas_token = sas_token
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._queue_cache = dict()
        self._aqueue_cache = dict()
        self._op_converter = OperationConverter()
        self._result_converter = ResultConverter()

        AzureProvider.__init__(
            self,
            credential_type=credential_type,
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
            certificate_path=certificate_path,
            **kwargs,
        )

    def __supports__(self, feature: str) -> bool:
        return feature not in [
            QueueFeature.BUILTIN_DLQ,
            QueueFeature.CONFIGURABLE_DLQ,
        ]

    def __setup__(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from azure.storage.queue import QueueServiceClient

        (self._credential, self._client) = self._init_credential_client(
            QueueServiceClient, self._get_credential
        )

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        from azure.storage.queue.aio import QueueServiceClient

        (
            self._acredential,
            self._aclient,
        ) = self._init_credential_client(
            QueueServiceClient, self._aget_credential
        )

    def _init_credential_client(
        self,
        service_client,
        get_credential,
    ):
        credential = None
        if self.connection_string is not None:
            client = service_client.from_connection_string(
                self.connection_string,
                **self.nparams,
            )
        elif self.account_key is not None:
            client = service_client(
                account_url=self.account_url,
                credential=self.account_key,
                **self.nparams,
            )
        elif self.sas_token is not None:
            client = service_client(
                account_url=self.account_url,
                credential=self.account_key,
                **self.nparams,
            )
        else:
            credential = get_credential()
            client = service_client(
                account_url=self.account_url,
                credential=credential,
                **self.nparams,
            )
        return credential, client

    def _get_queue_name(
        self, op_parser: MessagingOperationParser
    ) -> str | None:
        queue_name = op_parser.get_queue()
        queue_name = queue_name or self.queue or self.__component__.queue
        return queue_name

    def _get_queue(
        self, op_parser: MessagingOperationParser
    ) -> AzureQueue | None:
        if op_parser.is_resource_op():
            return None
        queue_name = self._get_queue_name(op_parser)
        if queue_name is None:
            raise BadRequestError("Queue name must be specified")
        if queue_name in self._queue_cache:
            return self._queue_cache[queue_name]

        client = self._client.get_queue_client(queue_name)
        queue = AzureQueue(
            client=client,
            helper=ClientHelper(client, self._op_converter),
        )
        self._queue_cache[queue_name] = queue
        return queue

    def _aget_queue(
        self, op_parser: MessagingOperationParser
    ) -> AzureQueue | None:
        if op_parser.is_resource_op():
            return None
        queue_name = self._get_queue_name(op_parser)
        if queue_name is None:
            raise BadRequestError("Queue name must be specified")
        if queue_name in self._aqueue_cache:
            return self._aqueue_cache[queue_name]
        client = self._aclient.get_queue_client(queue_name)
        queue = AzureQueue(
            client=client,
            helper=AsyncClientHelper(client, self._op_converter),
        )
        self._aqueue_cache[queue_name] = queue
        return queue

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = MessagingOperationParser(operation)
        queue = None
        resource_helper = None
        if not op_parser.is_resource_op():
            queue = self._get_queue(op_parser)
        resource_helper = ResourceHelper(
            self._client,
            self._queue_cache,
        )
        ncall = self._get_ncall(
            op_parser,
            queue,
            self._client,
            resource_helper,
        )
        if ncall is None:
            return super().__run__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.__asetup__(context=context)
        op_parser = MessagingOperationParser(operation)
        queue = None
        resource_helper = None
        if not op_parser.is_resource_op():
            queue = self._aget_queue(op_parser)
        resource_helper = AsyncResourceHelper(
            self._aclient,
            self._aqueue_cache,
        )
        ncall = self._get_ncall(
            op_parser,
            queue,
            self._aclient,
            resource_helper,
        )
        if ncall is None:
            return await super().__arun__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(nresult, op_parser)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: MessagingOperationParser,
        queue: AzureQueue | None,
        service_client: Any,
        resource_helper: Any,
    ) -> NCall | None:
        if queue is not None:
            queue_client = queue.client
            helper = queue.helper
        op_converter = self._op_converter
        call = None
        nargs = op_parser.get_nargs()

        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            args = op_converter.convert_create_queue(
                queue_name=self._get_queue_name(op_parser),
                config=op_parser.get_queue_config(),
            )
            where_exists = op_parser.get_where_exists()
            call = NCall(
                service_client.create_queue,
                args,
                nargs,
                {
                    ResourceExistsError: (
                        ConflictError if where_exists is False else None
                    )
                },
            )
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            args = {
                "queue_name": self._get_queue_name(op_parser),
                "config": op_parser.get_queue_config(),
                "nargs": nargs,
            }
            call = NCall(
                resource_helper.update_queue,
                args,
                nargs,
            )
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            args = {
                "queue_name": self._get_queue_name(op_parser),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            where_exists = op_parser.get_where_exists()
            call = NCall(
                resource_helper.drop_queue,
                args,
                nargs,
                {
                    ResourceNotFoundError: (
                        NotFoundError if where_exists is True else None
                    )
                },
            )
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            args = {
                "queue_name": self._get_queue_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                resource_helper.has_queue,
                args,
                nargs,
            )
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            args = {
                "queue_name": self._get_queue_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                resource_helper.get_queue,
                args,
                nargs,
            )
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_queues, args)
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            args = op_converter.convert_put(
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            call = NCall(
                queue_client.send_message,
                args,
                nargs,
            )
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            args = {
                "batch": op_parser.get_batch(),
                "nargs": nargs,
            }
            call = NCall(
                helper.batch,
                args,
                nargs,
            )
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            args = {
                "config": op_parser.get_pull_config(),
                "nargs": nargs,
            }
            call = NCall(
                helper.pull,
                args,
                nargs,
            )
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            args = op_converter.convert_ack(op_parser.get_key())
            call = NCall(
                queue_client.delete_message,
                args,
                nargs,
            )
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            args = op_converter.convert_nack(
                op_parser.get_key(),
            )
            call = NCall(
                queue_client.update_message,
                args,
                nargs,
            )
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            args = op_converter.convert_extend(
                op_parser.get_key(),
                op_parser.get_timeout(),
            )
            call = NCall(
                queue_client.update_message,
                args,
                nargs,
            )
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            call = NCall(
                queue_client.clear_messages,
                None,
                nargs,
            )
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            args = {
                "queue_name": op_parser.get_queue(),
                "nargs": nargs,
            }
            call = NCall(resource_helper.close, args)
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: MessagingOperationParser,
    ) -> Any:
        result_converter = ResultConverter()
        result: Any = None
        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            result = None
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            result = None
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            result = None
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            result = nresult
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            result = nresult
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            result = nresult
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            result = None
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            result = None
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            result = result_converter.convert_pull(nresult)
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            result = None
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            result = None
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            result = result_converter.convert_extend(nresult)
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            result = None
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            result = None
        return result


class ClientHelper:
    client: Any
    op_converter: OperationConverter

    def __init__(self, client: Any, op_converter: OperationConverter):
        self.client = client
        self.op_converter = op_converter

    def batch(self, batch: MessageBatch, nargs: Any) -> None:
        for operation in batch.operations:
            op_parser = MessagingOperationParser(operation)
            args = self.op_converter.convert_put(
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            NCall(
                self.client.send_message,
                args,
                nargs,
            ).invoke()

    def pull(self, config: MessagePullConfig | None, nargs: Any) -> list:
        args = self.op_converter.convert_pull(config)
        nresult = NCall(
            self.client.receive_messages,
            args,
            nargs,
        ).invoke()
        return nresult


class AsyncClientHelper:
    client: Any
    op_converter: OperationConverter

    def __init__(self, client: Any, op_converter: OperationConverter):
        self.client = client
        self.op_converter = op_converter

    async def batch(self, batch: MessageBatch, nargs: Any) -> None:
        for operation in batch.operations:
            op_parser = MessagingOperationParser(operation)
            args = self.op_converter.convert_put(
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            await NCall(
                self.client.send_message,
                args,
                nargs,
            ).ainvoke()

    async def pull(self, config: MessagePullConfig | None, nargs: Any) -> list:
        args = self.op_converter.convert_pull(config)
        response = NCall(
            self.client.receive_messages,
            args,
            nargs,
        ).invoke()
        nresult = []
        async for message in response:
            nresult.append(message)
        return nresult


class ResourceHelper:
    client: Any
    queue_cache: dict[str, AzureQueue]

    def __init__(self, client: Any, queue_cache: dict):
        self.client = client
        self.queue_cache = queue_cache

    def update_queue(
        self,
        queue_name: str,
        config: QueueConfig | None,
        nargs: Any,
    ) -> None:
        try:
            queue_client = self.client.get_queue_client(queue_name)
            queue_client.get_queue_properties()
        except ResourceNotFoundError:
            raise NotFoundError

    def drop_queue(
        self,
        queue_name: str,
        where_exists: bool,
        nargs: Any,
    ) -> None:
        if where_exists:
            try:
                queue_client = self.client.get_queue_client(queue_name)
                queue_client.get_queue_properties()
            except ResourceNotFoundError:
                raise NotFoundError
        NCall(
            self.client.delete_queue,
            {"queue": queue_name},
            nargs,
        ).invoke()

    def has_queue(
        self,
        queue_name: str,
        nargs: Any,
    ) -> bool:
        try:
            queue_client = self.client.get_queue_client(queue_name)
            queue_client.get_queue_properties()
            return True
        except ResourceNotFoundError:
            return False

    def get_queue(
        self,
        queue_name: str,
        nargs: Any,
    ) -> QueueInfo:
        try:
            queue_client = self.client.get_queue_client(queue_name)
            qp: QueueProperties = queue_client.get_queue_properties()
            return QueueInfo(
                name=queue_name,
                active_message_count=qp.approximate_message_count,
                nref=qp.name,
                config=QueueConfig(
                    nconfig=qp.metadata,
                ),
            )
        except ResourceNotFoundError:
            raise NotFoundError

    def list_queues(self, nargs: Any) -> list[str]:
        queues = NCall(self.client.list_queues, None, nargs).invoke()
        return [queue.name for queue in queues]

    def close(self, queue_name: str | None, nargs: Any) -> None:
        if queue_name is not None:
            if queue_name in self.queue_cache:
                queue = self.queue_cache[queue_name]
                queue.client.close()
        else:
            for queue in self.queue_cache.values():
                queue.client.close()
            self.queue_cache.clear()
            self.client.close()


class AsyncResourceHelper:
    client: Any
    queue_cache: dict[str, AzureQueue]

    def __init__(self, client: Any, queue_cache: dict):
        self.client = client
        self.queue_cache = queue_cache

    async def update_queue(
        self,
        queue_name: str,
        config: QueueConfig | None,
        nargs: Any,
    ) -> None:
        try:
            queue_client = self.client.get_queue_client(queue_name)
            await queue_client.get_queue_properties()
        except ResourceNotFoundError:
            raise NotFoundError

    async def drop_queue(
        self,
        queue_name: str,
        where_exists: bool,
        nargs: Any,
    ) -> None:
        if where_exists:
            try:
                queue_client = self.client.get_queue_client(queue_name)
                await queue_client.get_queue_properties()
            except ResourceNotFoundError:
                raise NotFoundError
        await NCall(
            self.client.delete_queue,
            {"queue": queue_name},
            nargs,
        ).ainvoke()

    async def has_queue(
        self,
        queue_name: str,
        nargs: Any,
    ) -> bool:
        try:
            queue_client = self.client.get_queue_client(queue_name)
            await queue_client.get_queue_properties()
            return True
        except ResourceNotFoundError:
            return False

    async def get_queue(
        self,
        queue_name: str,
        nargs: Any,
    ) -> QueueInfo:
        try:
            queue_client = self.client.get_queue_client(queue_name)
            qp: QueueProperties = await queue_client.get_queue_properties()
            return QueueInfo(
                name=queue_name,
                active_message_count=qp.approximate_message_count,
                nref=qp.name,
                config=QueueConfig(
                    nconfig=qp.metadata,
                ),
            )
        except ResourceNotFoundError:
            raise NotFoundError

    async def list_queues(self, nargs: Any) -> list[str]:
        queues = NCall(self.client.list_queues, None, nargs).invoke()
        return [queue.name async for queue in queues]

    async def close(self, queue_name: str | None, nargs: Any) -> None:
        if queue_name is not None:
            if queue_name in self.queue_cache:
                queue = self.queue_cache[queue_name]
                await queue.client.close()
                self.queue_cache.pop(queue_name)
        else:
            for queue in self.queue_cache.values():
                await queue.client.close()
            self.queue_cache.clear()
            await self.client.close()


class OperationConverter:
    def convert_create_queue(
        self,
        queue_name: str | None,
        config: QueueConfig | None,
    ) -> dict:
        args: dict = {"name": queue_name}
        if config and config.nconfig:
            args = args | config.nconfig
        return args

    def convert_put(
        self,
        value: MessageValueType,
        metadata: dict | None,
        properties: MessageProperties | None,
        config: MessagePutConfig | None,
    ) -> dict:
        content_type = None
        visibility_timeout = 0
        body: str | None = None

        if isinstance(value, str):
            body = value
            content_type = "text/plain"
        elif isinstance(value, bytes):
            body = base64.b64encode(value).decode("utf-8")
            content_type = "application/octet-stream"
        elif isinstance(value, dict):
            body = json.dumps(value)
            content_type = "application/json"
        elif isinstance(value, DataModel):
            body = value.to_json()
            content_type = "application/json"
        else:
            raise BadRequestError("Message type not supported")

        content: dict = {
            "body": body,
            "content_type": content_type,
        }

        if metadata:
            content["metadata"] = metadata
        if properties:
            if properties.message_id:
                content["message_id"] = properties.message_id
            if properties.group_id:
                content["group_id"] = properties.group_id

        if config is not None and config.delay is not None:
            visibility_timeout = config.delay

        return {
            "content": json.dumps(content),
            "visibility_timeout": visibility_timeout,
        }

    def convert_pull(self, config: MessagePullConfig | None) -> dict:
        args: dict = {}
        if config and config.max_count is not None:
            args["messages_per_page"] = config.max_count
        else:
            args["messages_per_page"] = 1
        if config and config.visibility_timeout:
            args["visibility_timeout"] = config.visibility_timeout
        return args

    def convert_ack(self, key: MessageKey) -> dict:
        message_data = key.nref
        return {
            "message": message_data["id"],
            "pop_receipt": message_data["pop_receipt"],
        }

    def convert_nack(self, key: MessageKey) -> dict:
        message_data = key.nref
        return {
            "message": message_data["id"],
            "pop_receipt": message_data["pop_receipt"],
            "visibility_timeout": 0,
            "content": message_data["content"],
        }

    def convert_extend(self, key: MessageKey, timeout: int | None) -> dict:
        message_data = key.nref
        return {
            "message": message_data["id"],
            "pop_receipt": message_data["pop_receipt"],
            "visibility_timeout": timeout,
            "content": message_data["content"],
        }


class ResultConverter:
    def _convert_key(self, message: QueueMessage) -> MessageKey:
        return MessageKey(
            id=message.id,
            nref={
                "id": message.id,
                "pop_receipt": message.pop_receipt,
                "content": message.content,
            },
        )

    def _convert_value(self, content: dict) -> MessageValueType:
        content_type = "text/plain"
        content_type = content.get("content_type", "text/plain")
        if content_type == "text/plain":
            return content.get("body", {})
        elif content_type == "application/octet-stream":
            return base64.b64decode(content.get("body", {}))
        elif content_type == "application/json":
            return json.loads(content.get("body", {}))
        return content.get("body", {})

    def _convert_metadata(self, content: dict) -> dict | None:
        if "metadata" in content:
            return content["metadata"]
        return None

    def _convert_properties(
        self,
        message: QueueMessage,
        content: dict,
    ) -> MessageProperties:
        return MessageProperties(
            message_id=content.get("message_id", None),
            group_id=content.get("group_id", None),
            content_type=content.get("content_type", "text/plain"),
            enqueued_time=(
                message.inserted_on.timestamp()
                if message.inserted_on
                else None
            ),
            delivery_count=message.dequeue_count,
        )

    def _convert_message(self, message: QueueMessage) -> MessageItem:
        content = json.loads(message.content)
        return MessageItem(
            key=self._convert_key(message=message),
            value=self._convert_value(content=content),
            metadata=self._convert_metadata(content=content),
            properties=self._convert_properties(
                message=message, content=content
            ),
        )

    def convert_pull(self, messages: list[QueueMessage]) -> list[MessageItem]:
        result = []
        for message in messages:
            result.append(self._convert_message(message))
        return result

    def convert_extend(self, message: QueueMessage) -> MessageItem:
        return MessageItem(key=self._convert_key(message=message))


class AzureQueue:
    client: Any
    helper: Any

    def __init__(self, client: Any, helper: Any):
        self.client = client
        self.helper = helper



"""
Live Stream processed with av.
"""

__all__ = ["AV"]

from io import BytesIO
from typing import Any

import av
import av.error
import av.option
from x8.content.audio import Audio
from x8.content.image import Image
from x8.content.video import VideoAudio, VideoFrame
from x8.core import Context, Provider
from x8.core.exceptions import BadRequestError

from .._models import LiveStreamInfo


class AV(Provider):
    get_playlists: bool
    get_segments: bool

    _container: Any
    _video_stream: Any
    _audio_stream: Any
    _stream_id: Any
    _playlist_url: str | None

    def __init__(
        self,
        get_playlists: bool = True,
        get_segments: bool = False,
        **kwargs,
    ):
        """_summary_

        Args:
            get_playlists:
                A value indicating whether to directly parse
                the m3u8 master list to get the playlists and choose the
                best stream instead of passing the url to directly to av,
                which might load slowly.
                Defaults to True.
            get_segments:
                A value indicating whether to directly parse
                the m3u8 playlist to get the segments and
                pass the last segment to av.
                Defaults to True.
        """
        self.get_playlists = get_playlists
        self.get_segments = get_segments

        self._container = None
        self._video_stream = None
        self._audio_stream = None
        self._stream_id = None
        self._playlist_url = None

    def __setup__(self, context: Context | None = None) -> None:
        if self._container is not None:
            return
        av_url = self.__component__.source
        if self._playlist_url is not None:
            av_url = self._playlist_url
        else:
            if self.get_playlists:
                av_url = self._get_playlist_url(
                    self.__component__.source,
                    self.__component__.min_resolution,
                )
                self._playlist_url = av_url

        if self.get_segments:
            av_url = self._get_segment_url(av_url)

        self._container = av.open(av_url)
        if not self.get_playlists and not self.get_segments:
            self._init_best_stream_from_container(
                self.__component__.min_resolution
            )
        else:
            self._video_stream = self._container.streams.video[0]
            self._audio_stream = self._container.streams.audio[0]
            self._stream_id = 0

    def get_info(self, **kwargs) -> LiveStreamInfo:
        return LiveStreamInfo(
            width=self._video_stream.width,
            height=self._video_stream.height,
            total_frames=self._video_stream.frames,
            frame_rate=self._video_stream.average_rate,
            format=(
                self._video_stream.format.name
                if self._video_stream.format
                else None
            ),
            codec=self._video_stream.codec_context.name,
            pixel_format=self._video_stream.pix_fmt,
            start_time=(
                self._video_stream.start_time * self._video_stream.time_base
                if self._video_stream.start_time
                else None
            ),
        )

    def get_frame(
        self,
        **kwargs,
    ) -> VideoFrame | None:
        if self._video_stream is None:
            raise BadRequestError("No video stream found.")
        try:
            for frame in self._container.decode(video=self._stream_id):
                frame_timestamp = frame.pts * float(frame.time_base)
                frame_index = None
                if self._video_stream.average_rate is not None:
                    frame_index = int(
                        frame.pts
                        * frame.time_base
                        * self._video_stream.average_rate
                    )
                return VideoFrame(
                    timestamp=frame_timestamp,
                    frame_index=frame_index,
                    key_frame=frame.key_frame,
                    width=frame.width,
                    height=frame.height,
                    image=Image(_pil_image=frame.to_image()),
                )
            return None
        except av.error.EOFError:
            return None

    def get_audio(
        self,
        duration: float,
        format: str | None = "mp3",
        codec: str | None = None,
        rate: int | None = None,
        channels: int | None = None,
        **kwargs,
    ) -> VideoAudio | None:
        if self._audio_stream is None:
            raise BadRequestError("No audio stream found.")
        if format is not None:
            format = format.lower()
        output_buffer = BytesIO()
        output_container = av.open(output_buffer, mode="w", format=format)
        if format == "wav" and codec is None:
            codec = "pcm_s16le"
        elif codec is None:
            codec = format
        args: dict = {"codec_name": codec}
        if rate is not None:
            args["rate"] = rate
        if channels == 1:
            args["layout"] = "mono"
        output_stream: Any = output_container.add_stream(**args)

        start_timestamp = None
        end_timestamp = None
        done = False
        for packet in self._container.demux(self._audio_stream):
            for frame in packet.decode():
                if frame.pts is not None:
                    current_frame_timestamp = frame.pts * float(
                        frame.time_base
                    )
                    if start_timestamp is None:
                        start_timestamp = current_frame_timestamp
                    end_timestamp = current_frame_timestamp
                    if duration is not None:
                        if end_timestamp - start_timestamp >= duration:
                            done = True
                            break
                for out_packet in output_stream.encode(frame):
                    output_container.mux(out_packet)
            if done is True:
                break

        # Flush the encoder and close the output container
        for out_packet in output_stream.encode(None):
            output_container.mux(out_packet)
        output_container.close()

        output_buffer.seek(0)
        if start_timestamp is None and end_timestamp is None:
            return None

        return VideoAudio(
            start=start_timestamp,
            end=end_timestamp,
            format=format,
            codec=codec,
            audio=Audio(stream=output_buffer),
        )

    def close(self):
        if self._container is not None:
            self._container.close()
            self._container = None

    def _get_playlist_url(self, url: str, min_resolution: int) -> str:
        import urllib.parse

        import m3u8

        base_url = url.rsplit("/", 1)[0] + "/"
        selected_resolution = 100000
        selected_playlist: Any = None
        m3u8_obj = m3u8.load(url)
        for playlist in m3u8_obj.playlists:
            height = playlist.stream_info.resolution[1]
            if height >= min_resolution and height < selected_resolution:
                selected_playlist = playlist
                selected_resolution = height

        if selected_playlist is None:
            raise BadRequestError(
                "Video stream with specified min resolution not found."
            )
        return urllib.parse.urljoin(base_url, selected_playlist.uri)

    def _get_segment_url(self, url: str) -> str:
        import urllib.parse

        import m3u8

        base_url = url.rsplit("/", 1)[0] + "/"
        playlist_obj = m3u8.load(url)
        return urllib.parse.urljoin(base_url, playlist_obj.segments[-1].uri)

    def _init_best_stream_from_container(self, min_resolution: int):
        selected_resolution = 100000
        stream: Any
        for stream in self._container.streams.video:
            if (
                stream.height >= min_resolution
                and stream.height < selected_resolution
            ):
                self._video_stream = stream
                self._stream_id = stream.id
                selected_resolution = stream.height

        if self._stream_id is None:
            raise BadRequestError(
                "Video stream with specified min resolution not found."
            )

        for stream in self._container.streams.audio:
            if stream.id == self._stream_id:
                self._audio_stream = stream



from x8.core import Response
from x8.storage.secret_store import SecretItem, SecretList

from ._providers import get_component


class SecretStoreSyncAndAsyncClient:
    def __init__(self, provider_type: str, async_call: bool):
        self.client = get_component(provider_type)
        self.async_call = async_call
        self.provider_type = provider_type

    async def get(self, key, **kwargs) -> Response[SecretItem]:
        if self.async_call:
            return await self.client.aget(key=key, **kwargs)
        return self.client.get(key=key, **kwargs)

    async def get_metadata(self, key, **kwargs) -> Response[SecretItem]:
        if self.async_call:
            return await self.client.aget_metadata(key=key, **kwargs)
        return self.client.get_metadata(key=key, **kwargs)

    async def get_versions(self, key, **kwargs) -> Response[SecretItem]:
        if self.async_call:
            return await self.client.aget_versions(key=key, **kwargs)
        return self.client.get_versions(key=key, **kwargs)

    async def put(
        self, key, value, metadata=None, where=None, **kwargs
    ) -> Response[SecretItem]:
        if self.async_call:
            return await self.client.aput(
                key=key, value=value, metadata=metadata, where=where, **kwargs
            )
        return self.client.put(
            key=key, value=value, metadata=metadata, where=where, **kwargs
        )

    async def update(self, key, value, **kwargs) -> Response[SecretItem]:
        if self.async_call:
            return await self.client.aupdate(key=key, value=value, **kwargs)
        return self.client.update(key=key, value=value, **kwargs)

    async def update_metadata(
        self, key, metadata, **kwargs
    ) -> Response[SecretItem]:
        if self.async_call:
            return await self.client.aupdate_metadata(
                key=key, metadata=metadata, **kwargs
            )
        return self.client.update_metadata(
            key=key, metadata=metadata, **kwargs
        )

    async def delete(self, key, **kwargs) -> Response[None]:
        if self.async_call:
            return await self.client.adelete(key=key, **kwargs)
        return self.client.delete(key=key, **kwargs)

    async def query(self, where=None, **kwargs) -> Response[SecretList]:
        if self.async_call:
            return await self.client.aquery(where=where, **kwargs)
        return self.client.query(where=where, **kwargs)

    async def count(self, where=None, **kwargs) -> Response[int]:
        if self.async_call:
            return await self.client.acount(where=where, **kwargs)
        return self.client.count(where=where, **kwargs)

    async def close(self) -> Response[None]:
        if self.async_call:
            return await self.client.aclose()
        else:
            return Response(result=None)



# type: ignore

import time
from datetime import datetime, timezone

import pytest

from x8.messaging.pubsub import (
    ConflictError,
    MessageBatch,
    MessageItem,
    MessagePullConfig,
    NotFoundError,
    SubscriptionConfig,
    SubscriptionInfo,
    TopicConfig,
    TopicInfo,
)

from ._data import messages
from ._providers import PubSubProvider
from ._sync_and_async_client import PubSubSyncAndAsyncClient

except_config_providers = [
    PubSubProvider.REDIS,
]

except_count_providers = [
    PubSubProvider.GOOGLE_PUBSUB,
]

except_pull_visibility_providers = [
    PubSubProvider.AZURE_SERVICE_BUS,
    PubSubProvider.GOOGLE_PUBSUB,
    PubSubProvider.REDIS,
]

topic_name = "topictest"
subscription_names = ["sub1", "sub2"]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        PubSubProvider.AMAZON_SNS,
        PubSubProvider.AZURE_SERVICE_BUS,
        PubSubProvider.GOOGLE_PUBSUB,
        PubSubProvider.REDIS,
        PubSubProvider.POSTGRESQL,
        PubSubProvider.SQLITE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_create_drop(provider_type: str, async_call: bool):
    new_topic = f"ntest{str(async_call).lower()}"
    sub1 = f"nsub1{str(async_call).lower()}"
    sub2 = f"nsub2{str(async_call).lower()}"
    topic_config = TopicConfig(ttl=600)
    new_topic_config = TopicConfig(ttl=900)
    subscription_config = SubscriptionConfig(
        visibility_timeout=10,
        ttl=600,
    )
    new_subscription_config = SubscriptionConfig(
        visibility_timeout=20,
        ttl=900,
    )
    client = PubSubSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    response = await client.list_topics()
    if new_topic in response.result:
        await client.drop_topic(topic=new_topic)

    response = await client.list_topics()
    assert new_topic not in response.result

    response = await client.has_topic(topic=new_topic)
    assert not response.result

    with pytest.raises(NotFoundError):
        await client.get_topic(topic=new_topic)

    await client.create_topic(
        topic=new_topic,
        config=topic_config,
    )

    response = await client.list_topics()
    assert new_topic in response.result

    response = await client.has_topic(topic=new_topic)
    assert response.result

    response = await client.get_topic(topic=new_topic)
    assert_topic_info(
        result=response.result,
        topic=new_topic,
        config=topic_config,
    )

    await client.update_topic(
        topic=new_topic,
        config=new_topic_config,
    )
    response = await client.get_topic(topic=new_topic)
    assert_topic_info(
        result=response.result,
        topic=new_topic,
        config=new_topic_config,
    )

    await client.create_topic(
        topic=new_topic,
    )

    with pytest.raises(ConflictError):
        await client.create_topic(
            topic=new_topic,
            where="not_exists()",
        )

    response = await client.list_subscriptions(topic=new_topic)
    assert sub1 not in response.result
    assert sub2 not in response.result

    response = await client.has_subscription(
        topic=new_topic,
        subscription=sub1,
    )
    assert not response.result

    with pytest.raises(NotFoundError):
        await client.get_subscription(
            topic=new_topic,
            subscription=sub1,
        )

    await client.create_subscription(
        topic=new_topic, subscription=sub1, config=subscription_config
    )
    await client.create_subscription(
        topic=new_topic, subscription=sub2, config=subscription_config
    )
    wait_for_create_subscription(provider_type)
    response = await client.list_subscriptions(topic=new_topic)
    assert sub1 in response.result
    assert sub2 in response.result

    response = await client.has_subscription(
        topic=new_topic,
        subscription=sub1,
    )
    assert response.result

    response = await client.get_subscription(
        topic=new_topic,
        subscription=sub1,
    )
    assert_subscription_info(
        provider_type=provider_type,
        result=response.result,
        topic=new_topic,
        subscription=sub1,
        config=subscription_config,
        count=0,
    )

    await client.update_subscription(
        topic=new_topic,
        subscription=sub1,
        config=new_subscription_config,
    )
    response = await client.get_subscription(
        topic=new_topic,
        subscription=sub1,
    )
    assert_subscription_info(
        provider_type=provider_type,
        result=response.result,
        topic=new_topic,
        subscription=sub1,
        config=new_subscription_config,
        count=0,
    )

    await client.create_subscription(topic=new_topic, subscription=sub1)
    with pytest.raises(ConflictError):
        await client.create_subscription(
            topic=new_topic, subscription=sub1, where="not_exists()"
        )

    await client.drop_subscription(topic=new_topic, subscription=sub1)
    response = await client.list_subscriptions(topic=new_topic)
    assert sub1 not in response.result
    assert sub2 in response.result

    await client.drop_subscription(topic=new_topic, subscription=sub1)
    with pytest.raises(NotFoundError):
        await client.drop_subscription(
            topic=new_topic, subscription=sub1, where="exists()"
        )

    await client.drop_subscription(topic=new_topic, subscription=sub2)

    await client.drop_topic(topic=new_topic)

    response = await client.list_topics()
    assert new_topic not in response.result

    await client.drop_topic(topic=new_topic)

    with pytest.raises(NotFoundError):
        await client.drop_topic(
            topic=new_topic,
            where="exists()",
        )

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        PubSubProvider.AMAZON_SNS,
        PubSubProvider.AZURE_SERVICE_BUS,
        PubSubProvider.GOOGLE_PUBSUB,
        PubSubProvider.REDIS,
        PubSubProvider.POSTGRESQL,
        PubSubProvider.SQLITE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_put_pull(provider_type: str, async_call: bool):
    client = PubSubSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_topic_if_needed(provider_type, client)
    for subscription_name in subscription_names:
        await client.purge(
            config=dict(max_count=1, max_wait_time=2),
            topic=topic_name,
            subscription=subscription_name,
        )

    for subscription_name in subscription_names:
        res = await client.pull(
            config=dict(max_wait_time=1),
            topic=topic_name,
            subscription=subscription_name,
        )
        assert len(res.result) == 0

    for message in messages:
        await client.put(**message, topic=topic_name)
        for subscription_name in subscription_names:
            res = await client.pull(
                config=dict(max_wait_time=5),
                topic=topic_name,
                subscription=subscription_name,
            )
            assert_message(res.result[0], message)
            await client.ack(
                key=res.result[0].key,
                topic=topic_name,
                subscription=subscription_name,
            )

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        PubSubProvider.AMAZON_SNS,
        PubSubProvider.AZURE_SERVICE_BUS,
        PubSubProvider.GOOGLE_PUBSUB,
        PubSubProvider.REDIS,
        PubSubProvider.POSTGRESQL,
        PubSubProvider.SQLITE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_batch(provider_type: str, async_call: bool):
    client = PubSubSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_topic_if_needed(provider_type, client)
    for subscription_name in subscription_names:
        await client.purge(
            config=dict(max_count=1, max_wait_time=2),
            topic=topic_name,
            subscription=subscription_name,
        )

    batch = MessageBatch()
    for message in messages:
        batch.put(**message)
    await client.batch(batch=batch, topic=topic_name)

    for subscription_name in subscription_names:
        result = []
        while True:
            res = await client.pull(
                topic=topic_name,
                subscription=subscription_name,
                config=MessagePullConfig(max_wait_time=5),
            )
            for message in res.result:
                result.append(message)
                # If two messages have the same group_id,
                # (ordering_key in Google PubSub)
                # the subsequent message will not be delivered
                # until the first message is acked.
                await client.ack(
                    key=message.key,
                    topic=topic_name,
                    subscription=subscription_name,
                )
            if len(result) == len(messages):
                break
        assert len(result) == len(messages)
        assert_batch(result, messages)

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        PubSubProvider.AMAZON_SNS,
        PubSubProvider.AZURE_SERVICE_BUS,
        PubSubProvider.GOOGLE_PUBSUB,
        PubSubProvider.REDIS,
        PubSubProvider.POSTGRESQL,
        PubSubProvider.SQLITE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_nack(provider_type: str, async_call: bool):
    client = PubSubSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_topic_if_needed(provider_type, client)
    for subscription_name in subscription_names:
        await client.purge(
            config=dict(max_count=1, max_wait_time=2),
            topic=topic_name,
            subscription=subscription_name,
        )

    message = messages[0]
    await client.put(
        **message,
        topic=topic_name,
    )
    for subscription_name in subscription_names:
        res = await client.pull(
            topic=topic_name,
            subscription=subscription_name,
        )
        assert_message(res.result[0], message)
        await client.nack(
            key=res.result[0].key,
            topic=topic_name,
            subscription=subscription_name,
        )
        res = await client.pull(
            topic=topic_name,
            subscription=subscription_name,
        )
        assert_message(res.result[0], message)
        await client.extend(
            key=res.result[0].key,
            timeout=10,
            topic=topic_name,
            subscription=subscription_name,
        )
        await client.ack(
            key=res.result[0].key,
            topic=topic_name,
            subscription=subscription_name,
        )
    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        PubSubProvider.AMAZON_SNS,
        PubSubProvider.AZURE_SERVICE_BUS,
        PubSubProvider.GOOGLE_PUBSUB,
        PubSubProvider.REDIS,
        PubSubProvider.POSTGRESQL,
        PubSubProvider.SQLITE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_count(provider_type: str, async_call: bool):
    if provider_type in except_count_providers:
        pytest.skip("Not supported for this provider")
    client = PubSubSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_topic_if_needed(provider_type, client)
    for subscription_name in subscription_names:
        await client.purge(
            config=dict(max_count=1, max_wait_time=2),
            topic=topic_name,
            subscription=subscription_name,
        )

    for message in messages:
        await client.put(**message, topic=topic_name)

    for subscription_name in subscription_names:
        res = await client.get_subscription(
            topic=topic_name, subscription=subscription_name
        )
        assert_count(res.result, len(messages))
        for message in messages:
            res = await client.pull(
                topic=topic_name, subscription=subscription_name
            )
            await client.ack(
                key=res.result[0].key,
                topic=topic_name,
                subscription=subscription_name,
            )

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        PubSubProvider.AMAZON_SNS,
        PubSubProvider.AZURE_SERVICE_BUS,
        PubSubProvider.GOOGLE_PUBSUB,
        PubSubProvider.REDIS,
        PubSubProvider.POSTGRESQL,
        PubSubProvider.SQLITE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_visibility(provider_type: str, async_call: bool):
    if provider_type in except_pull_visibility_providers:
        pytest.skip("Not supported for this provider")
    client = PubSubSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_topic_if_needed(provider_type, client)
    for subscription_name in subscription_names:
        await client.purge(
            config=dict(max_count=1, max_wait_time=2),
            topic=topic_name,
            subscription=subscription_name,
        )

    message = messages[0]
    await client.put(**message, topic=topic_name)
    for subscription_name in subscription_names:
        res = await client.pull(
            topic=topic_name,
            subscription=subscription_name,
            config=MessagePullConfig(visibility_timeout=1),
        )
        time.sleep(2)
        res = await client.pull(
            topic=topic_name,
            subscription=subscription_name,
            config=MessagePullConfig(visibility_timeout=10),
        )
        await client.ack(
            key=res.result[0].key,
            topic=topic_name,
            subscription=subscription_name,
        )
    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        PubSubProvider.AMAZON_SNS,
        PubSubProvider.AZURE_SERVICE_BUS,
        PubSubProvider.GOOGLE_PUBSUB,
        PubSubProvider.REDIS,
        PubSubProvider.POSTGRESQL,
        PubSubProvider.SQLITE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_close(provider_type: str, async_call: bool):
    client = PubSubSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_topic_if_needed(provider_type, client)
    for subscription_name in subscription_names:
        await client.purge(
            config=dict(max_count=1, max_wait_time=2),
            topic=topic_name,
            subscription=subscription_name,
        )
    message = messages[0]
    await client.put(**message)
    for subscription_name in subscription_names:
        res = await client.pull(
            topic=topic_name,
            subscription=subscription_name,
        )
        assert_message(res.result[0], message)
        await client.ack(
            key=res.result[0].key,
            topic=topic_name,
            subscription=subscription_name,
        )
    await client.close(topic=topic_name)

    await client.put(**message)
    for subscription_name in subscription_names:
        res = await client.pull(
            topic=topic_name,
            subscription=subscription_name,
        )
        assert_message(res.result[0], message)
        await client.ack(
            key=res.result[0].key,
            topic=topic_name,
            subscription=subscription_name,
        )
    await client.close()


def wait_for_create_subscription(provider_type: str):
    if provider_type == PubSubProvider.AMAZON_SNS:
        time.sleep(5)


def assert_count(result: SubscriptionInfo, count: int):
    assert result.active_message_count >= count


def assert_topic_info(
    result: TopicInfo,
    topic: str,
    config: TopicConfig,
):
    assert result.name == topic
    assert result.nref is not None
    # assert result.config.ttl == config.ttl


def assert_subscription_info(
    provider_type: str,
    result: SubscriptionInfo,
    topic: str,
    subscription: str,
    config: SubscriptionConfig,
    count: int,
):
    assert result.name == subscription
    assert result.topic == topic
    if provider_type not in except_config_providers:
        assert result.config.visibility_timeout == config.visibility_timeout
        assert result.config.ttl == config.ttl
    if provider_type not in except_count_providers:
        assert result.active_message_count == count
    assert result.nref is not None


def assert_batch(result: list[MessageItem], messages: list[dict]):
    for message in messages:
        found = False
        for rmessage in result:
            if (
                rmessage.properties.message_id
                == message["properties"]["message_id"]
            ):
                assert_message(rmessage, message)
                found = True
                break
        if not found:
            raise AssertionError("Message not found")


def assert_message(result: MessageItem, message: dict):
    assert result.value == message["value"]
    assert result.metadata == message["metadata"]
    assert (
        result.properties.content_type == message["properties"]["content_type"]
    )
    assert result.properties.message_id == message["properties"]["message_id"]
    assert result.properties.group_id == message["properties"]["group_id"]
    diff_time_seconds = (
        datetime.now(timezone.utc).timestamp()
        - result.properties.enqueued_time
    )
    assert diff_time_seconds < 100


async def create_topic_if_needed(
    provider_type: str,
    client: PubSubSyncAndAsyncClient,
):
    response = await client.list_topics()
    topics = response.result
    if topic_name not in topics:
        await client.create_topic(
            topic=topic_name, config=TopicConfig(ttl=600)
        )
    response = await client.list_subscriptions()
    subscriptions = response.result
    for subscription_name in subscription_names:
        if subscription_name not in subscriptions:
            await client.create_subscription(
                topic=topic_name,
                subscription=subscription_name,
                config=SubscriptionConfig(visibility_timeout=30),
            )



import asyncio
import base64
import time
from typing import Any, AsyncIterator, Iterator, List

import google.auth
import google.auth.transport.requests
import httpx
from x8._common.google_provider import GoogleProvider
from x8.content.image import ImageData
from x8.content.video import VideoData
from x8.core import Response
from x8.core.exceptions import BadRequestError, NotFoundError

from .._models import KeyFrame, Reference, VideoGenerationResult, VideoSize


class Google(GoogleProvider):
    project: str | None
    location: str
    model: str
    nparams: dict[str, Any] | None

    _credentials: Any
    _generate_url: str
    _poll_url: str
    _full_name_prefix: str

    def __init__(
        self,
        project: str | None = None,
        location: str = "us-central1",
        model: str = "veo-3.1-fast-generate-preview",
        service_account_info: str | None = None,
        service_account_file: str | None = None,
        access_token: str | None = None,
        nparams: dict[str, Any] | None = None,
        **kwargs,
    ):
        self.project = project
        self.location = location
        self.model = model
        self.nparams = nparams
        self._credentials = None
        self._generate_url = self._get_generate_url()
        self._poll_url = self._get_poll_url()
        self._full_name_prefix = self._get_full_name_prefix()
        super().__init__(
            service_account_info=service_account_info,
            service_account_file=service_account_file,
            access_token=access_token,
            **kwargs,
        )

    def _get_token(self) -> str:
        if not self._credentials:
            self._credentials = (
                self._get_credentials() or google.auth.default()
            )
            self._credentials, _ = google.auth.default()
            request = google.auth.transport.requests.Request()
            self._credentials.refresh(request)

        if self._credentials:
            return self._credentials.token

        raise BadRequestError("Unable to obtain access token")

    def generate(
        self,
        prompt: str | None = None,
        *,
        image: ImageData | None = None,
        references: List[Reference] | None = None,
        key_frames: List[KeyFrame] | None = None,
        negative_prompt: str | None = None,
        audio: bool | None = True,
        duration: float | None = None,
        size: VideoSize | None = None,
        samples: int | None = None,
        seed: int | None = None,
        quality: str | None = None,
        poll: bool = False,
        poll_interval: float = 5.0,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[VideoGenerationResult]
        | Iterator[Response[VideoGenerationResult]]
    ):
        body = self._convert_generate_args(
            prompt=prompt,
            image=image,
            references=references,
            key_frames=key_frames,
            negative_prompt=negative_prompt,
            audio=audio,
            duration=duration,
            size=size,
            samples=samples,
            seed=seed,
            quality=quality,
            nconfig=nconfig,
            **kwargs,
        )
        headers = self._get_headers()
        res = httpx.post(self._generate_url, json=body, headers=headers).json()
        if res.get("error"):
            return Response(
                result=VideoGenerationResult(
                    id="",
                    status="failed",
                    error=res["error"].get("message", "Unknown error"),
                )
            )

        id = res.get("name").split("/")[-1]

        if stream:

            def _poll_iter() -> Iterator[Response[VideoGenerationResult]]:
                yield Response(
                    result=VideoGenerationResult(id=id, status="queued")
                )

                operation_name = res["name"]
                poll_body = {"operationName": operation_name}

                while True:
                    poll_res = httpx.post(
                        self._poll_url, headers=headers, json=poll_body
                    ).json()
                    if poll_res.get("error"):
                        yield Response(
                            result=VideoGenerationResult(
                                id=id,
                                status="failed",
                                error=res["error"].get(
                                    "message", "Unknown error"
                                ),
                            )
                        )
                        return

                    if poll_res.get("done"):
                        videos = self._parse_videos_response(
                            poll_res["response"]
                        )
                        yield Response(
                            result=VideoGenerationResult(
                                id=id,
                                status="completed",
                                videos=videos,
                            )
                        )
                        return
                    yield Response(
                        result=VideoGenerationResult(
                            id=id,
                            status="in_progress",
                        )
                    )
                    time.sleep(poll_interval)

            return _poll_iter()

        if poll:
            operation_name = res["name"]
            poll_body = {"operationName": operation_name}
            while True:
                poll_res = httpx.post(
                    self._poll_url, headers=headers, json=poll_body
                ).json()
                if poll_res.get("error"):
                    return Response(
                        result=VideoGenerationResult(
                            id=id,
                            status="failed",
                            error=res["error"].get("message", "Unknown error"),
                        )
                    )
                if poll_res.get("done"):
                    return Response(
                        result=VideoGenerationResult(
                            id=id,
                            status="completed",
                            videos=self._parse_videos_response(
                                poll_res["response"]
                            ),
                        )
                    )
                time.sleep(poll_interval)
        return Response(result=VideoGenerationResult(id=id, status="queued"))

    def get(self, id: str, **kwargs: Any) -> Response[VideoGenerationResult]:
        body = {"operationName": self._get_full_name(id)}
        print(body)
        headers = self._get_headers()
        res = httpx.post(self._get_poll_url(), headers=headers, json=body)
        if res.status_code == 404:
            raise NotFoundError(f"Video generation with ID {id} not found.")
        res_json = res.json()
        operation_done = res_json.get("done", False)
        if not operation_done:
            return Response(
                result=VideoGenerationResult(
                    id=id,
                    status="in_progress",
                )
            )
        return Response(
            result=VideoGenerationResult(
                id=id,
                status="completed",
            )
        )

    def download(self, id: str, **kwargs: Any) -> Iterator[bytes]:
        body = {"operationName": self._get_full_name(id)}
        headers = self._get_headers()
        res = httpx.post(self._get_poll_url(), headers=headers, json=body)
        if res.status_code == 404:
            raise NotFoundError(f"Video generation with ID {id} not found.")
        res_json = res.json()
        if not res_json.get("done", False):
            raise NotFoundError(
                f"Video generation with ID {id} is not completed yet."
            )
        base_64_video = res_json["response"]["videos"][0]["bytesBase64Encoded"]
        video = base64.b64decode(base_64_video)
        return self._bytes_to_iterator(video)

    def list(self, **kwargs: Any) -> Response[List[VideoGenerationResult]]:
        return Response(result=[])

    def delete(self, id: str, **kwargs: Any) -> Response[None]:
        return Response(result=None)

    async def agenerate(
        self,
        prompt: str | None = None,
        *,
        image: ImageData | None = None,
        references: List[Reference] | None = None,
        key_frames: List[KeyFrame] | None = None,
        negative_prompt: str | None = None,
        audio: bool | None = True,
        duration: float | None = None,
        size: VideoSize | None = None,
        samples: int | None = None,
        seed: int | None = None,
        quality: str | None = None,
        poll: bool = False,
        poll_interval: float = 5.0,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[VideoGenerationResult]
        | AsyncIterator[Response[VideoGenerationResult]]
    ):
        body = self._convert_generate_args(
            prompt=prompt,
            image=image,
            references=references,
            key_frames=key_frames,
            negative_prompt=negative_prompt,
            audio=audio,
            duration=duration,
            size=size,
            samples=samples,
            seed=seed,
            quality=quality,
            nconfig=nconfig,
            **kwargs,
        )
        headers = self._get_headers()
        async with httpx.AsyncClient() as client:
            r = await client.post(
                self._generate_url, json=body, headers=headers
            )
            res = r.json()
        if res.get("error"):
            return Response(
                result=VideoGenerationResult(
                    id="",
                    status="failed",
                    error=res["error"].get("message", "Unknown error"),
                )
            )
        id = res.get("name").split("/")[-1]
        if stream:

            async def _poll_aiter() -> (
                AsyncIterator[Response[VideoGenerationResult]]
            ):
                yield Response(
                    result=VideoGenerationResult(id=id, status="queued")
                )
                operation_name = res["name"]
                poll_body = {"operationName": operation_name}

                while True:
                    async with httpx.AsyncClient() as client:
                        r = await client.post(
                            self._poll_url, headers=headers, json=poll_body
                        )
                        poll_res = r.json()
                    if poll_res.get("error"):
                        yield Response(
                            result=VideoGenerationResult(
                                id=id,
                                status="failed",
                                error=res["error"].get(
                                    "message", "Unknown error"
                                ),
                            )
                        )
                        return
                    if poll_res.get("done"):
                        videos = self._parse_videos_response(
                            poll_res["response"]
                        )
                        yield Response(
                            result=VideoGenerationResult(
                                id=id,
                                status="completed",
                                videos=videos,
                            )
                        )
                        return
                    yield Response(
                        result=VideoGenerationResult(
                            id=id,
                            status="in_progress",
                        )
                    )
                    await asyncio.sleep(poll_interval)

            return _poll_aiter()

        if poll:
            operation_name = res["name"]
            poll_body = {"operationName": operation_name}
            while True:
                async with httpx.AsyncClient() as client:
                    r = await client.post(
                        self._poll_url, headers=headers, json=poll_body
                    )
                    poll_res = r.json()
                if poll_res.get("error"):
                    return Response(
                        result=VideoGenerationResult(
                            id=id,
                            status="failed",
                            error=res["error"].get("message", "Unknown error"),
                        )
                    )
                if poll_res.get("done"):
                    res = Response(
                        result=VideoGenerationResult(
                            id=id,
                            status="completed",
                            videos=self._parse_videos_response(
                                poll_res["response"]
                            ),
                        )
                    )
                    return res
                time.sleep(poll_interval)
        return Response(result=VideoGenerationResult(id=id, status="queued"))

    async def aget(
        self, id: str, **kwargs: Any
    ) -> Response[VideoGenerationResult]:
        body = {"operationName": self._get_full_name(id)}
        print(body)
        headers = self._get_headers()
        async with httpx.AsyncClient() as client:
            res = await client.post(
                self._get_poll_url(), headers=headers, json=body
            )
        if res.status_code == 404:
            raise NotFoundError(f"Video generation with ID {id} not found.")
        res_json = res.json()
        operation_done = res_json.get("done", False)
        if not operation_done:
            return Response(
                result=VideoGenerationResult(
                    id=id,
                    status="in_progress",
                )
            )
        return Response(
            result=VideoGenerationResult(
                id=id,
                status="completed",
            )
        )

    async def alist(
        self, **kwargs: Any
    ) -> Response[List[VideoGenerationResult]]:
        return Response(result=[])

    async def adelete(self, id: str, **kwargs: Any) -> Response[None]:
        return Response(result=None)

    async def adownload(self, id: str, **kwargs: Any) -> AsyncIterator[bytes]:
        body = {"operationName": self._get_full_name(id)}
        headers = self._get_headers()
        async with httpx.AsyncClient() as client:
            res = await client.post(
                self._get_poll_url(), headers=headers, json=body
            )
        if res.status_code == 404:
            raise NotFoundError(f"Video generation with ID {id} not found.")
        res_json = res.json()
        if not res_json.get("done", False):
            raise NotFoundError(
                f"Video generation with ID {id} is not completed yet."
            )
        base_64_video = res_json["response"]["videos"][0]["bytesBase64Encoded"]
        video = base64.b64decode(base_64_video)
        return self._bytes_to_async_iterator(video)

    def _convert_generate_args(
        self,
        prompt: str | None = None,
        image: ImageData | None = None,
        references: List[Reference] | None = None,
        key_frames: List[KeyFrame] | None = None,
        negative_prompt: str | None = None,
        audio: bool | None = True,
        duration: float | None = None,
        size: VideoSize | None = None,
        samples: int | None = None,
        seed: int | None = None,
        quality: str | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        args: dict[str, Any] = {}
        instance: dict[str, Any] = {}
        parameters: dict[str, Any] = {}
        if prompt:
            instance["prompt"] = prompt
        if image:
            instance["image"] = self._get_image(image)
        if key_frames and len(key_frames) > 0:
            instance["lastFrame"] = self._get_image(key_frames[-1].image)
        if references:
            arg_references = []
            for reference in references:
                arg_reference: dict[str, Any] = {}
                if reference.type:
                    arg_reference["type"] = reference.type
                if reference.image:
                    arg_reference["image"] = self._get_image(reference.image)
                arg_references.append(arg_reference)
            instance["referenceImages"] = arg_references
        args["instances"] = [instance]
        args["parameters"] = parameters
        if not size:
            size = "1280x720"
        if size == "1280x720":
            parameters["resolution"] = "720p"
            parameters["aspectRatio"] = "16:9"
        elif size == "720x1280":
            parameters["resolution"] = "720p"
            parameters["aspectRatio"] = "9:16"
        elif size == "1920x1080":
            parameters["resolution"] = "1080p"
            parameters["aspectRatio"] = "16:9"
        elif size == "1080x1920":
            parameters["resolution"] = "1080p"
            parameters["aspectRatio"] = "9:16"
        if duration:
            parameters["duration"] = duration
        else:
            parameters["duration"] = 4
        if audio is not None:
            parameters["audio"] = audio
        if negative_prompt:
            parameters["negativePrompt"] = negative_prompt
        if samples:
            parameters["sampleCount"] = samples
        if seed:
            parameters["seed"] = seed
        if quality:
            parameters["compressionQuality"] = quality
        return args

    def _get_full_name(self, id: str) -> str:
        return f"{self._full_name_prefix}/{id}"

    def _get_generate_url(self) -> str:
        return f"{self._get_base_url()}:predictLongRunning"

    def _get_poll_url(self) -> str:
        return f"{self._get_base_url()}:fetchPredictOperation"

    def _get_base_url(self) -> str:
        project = self.project or self._get_default_project()
        return (
            f"https://{self.location}-aiplatform.googleapis.com/v1/"
            f"projects/{project}/locations/{self.location}/publishers/google/"
            f"models/{self.model}"
        )

    def _get_full_name_prefix(self) -> str:
        project = self.project or self._get_default_project()
        return (
            f"projects/{project}/locations/{self.location}/publishers/google/"
            f"models/{self.model}/operations"
        )

    def _get_headers(self) -> dict[str, str]:
        headers = {
            "Authorization": f"Bearer {self._get_token()}",
            "Content-Type": "application/json",
        }
        return headers

    def _get_image(self, image: ImageData) -> dict[str, Any]:
        args: dict[str, Any] = {}
        if image.source:
            args["gcrUri"] = image.source
        elif isinstance(image.content, str):
            args["bytesBase64Encoded"] = image.content
        elif isinstance(image.content, bytes):
            args["bytesBase64Encoded"] = base64.b64encode(
                image.content
            ).decode("utf-8")
        else:
            raise BadRequestError("Image data not provided.")
        if image.media_type:
            args["mimeType"] = image.media_type
        return args

    def _parse_videos_response(self, res: dict[str, Any]) -> List[VideoData]:
        results: List[VideoData] = []
        for video_data in res.get("videos", []):
            video = VideoData(
                content=video_data.get("bytesBase64Encoded"),
                media_type=video_data.get("mimeType"),
            )
            results.append(video)
        return results

    def _bytes_to_iterator(
        self, data: bytes, chunk_size: int = 1024
    ) -> Iterator[bytes]:
        for i in range(0, len(data), chunk_size):
            yield data[i : i + chunk_size]  # noqa: E203

    async def _bytes_to_async_iterator(
        self, data: bytes, chunk_size: int = 1024
    ) -> AsyncIterator[bytes]:
        for i in range(0, len(data), chunk_size):
            yield data[i : i + chunk_size]  # noqa: E203
            await asyncio.sleep(0)



# type: ignore
import asyncio
import os
import time

import pytest

from x8.storage.key_value_store import (
    KeyValueBatch,
    KeyValueItem,
    NotFoundError,
    PreconditionFailedError,
)

from ._data import batch_kvs, kvs, query_kvs
from ._providers import KeyValueStoreProvider
from ._sync_and_async_client import KeyValueStoreSyncAndAsyncClient

if os.name == "nt":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

providers = [
    KeyValueStoreProvider.REDIS,
    KeyValueStoreProvider.REDIS_SIMPLE,
    KeyValueStoreProvider.MEMCACHED,
    KeyValueStoreProvider.MEMORY,
    KeyValueStoreProvider.SQLITE,
    KeyValueStoreProvider.POSTGRESQL,
    KeyValueStoreProvider.DS_AMAZON_DYNAMODB,
    KeyValueStoreProvider.DS_AZURE_COSMOS_DB,
    KeyValueStoreProvider.DS_GOOGLE_FIRESTORE,
    KeyValueStoreProvider.DS_MONGODB,
    KeyValueStoreProvider.OS_AMAZON_S3,
    KeyValueStoreProvider.OS_AZURE_BLOB_STORAGE,
    KeyValueStoreProvider.OS_GOOGLE_CLOUD_STORAGE,
    KeyValueStoreProvider.OS_FILE_SYSTEM,
]

except_incr_float_providers = [
    KeyValueStoreProvider.MEMCACHED,
]
except_put_update_return_etag_providers = [
    KeyValueStoreProvider.MEMCACHED,
    KeyValueStoreProvider.REDIS_SIMPLE,
]
except_replace_etag_providers = [
    KeyValueStoreProvider.REDIS_SIMPLE,
]
except_delete_etag_providers = [
    KeyValueStoreProvider.MEMCACHED,
    KeyValueStoreProvider.REDIS_SIMPLE,
]
except_update_etag_providers = [
    KeyValueStoreProvider.MEMCACHED,
    KeyValueStoreProvider.REDIS_SIMPLE,
]
except_batch_get_etag_providers = [
    KeyValueStoreProvider.MEMCACHED,
    KeyValueStoreProvider.REDIS_SIMPLE,
]
except_get_etag_providers = [
    KeyValueStoreProvider.REDIS_SIMPLE,
]
except_update_string_providers = [
    KeyValueStoreProvider.REDIS,
    KeyValueStoreProvider.DS_AMAZON_DYNAMODB,
    KeyValueStoreProvider.DS_AZURE_COSMOS_DB,
    KeyValueStoreProvider.DS_GOOGLE_FIRESTORE,
    KeyValueStoreProvider.DS_MONGODB,
    KeyValueStoreProvider.OS_AMAZON_S3,
    KeyValueStoreProvider.OS_AZURE_BLOB_STORAGE,
    KeyValueStoreProvider.OS_GOOGLE_CLOUD_STORAGE,
    KeyValueStoreProvider.OS_FILE_SYSTEM,
]

queries = [
    {
        "where": None,
        "limit": None,
        "result_index": [0, 1, 2, 3, 4],
        "collection": "col1",
    },
    {
        "where": None,
        "limit": None,
        "result_index": [5, 6, 7, 8, 9, 10, 11, 12, 13],
        "collection": "col2",
    },
    {
        "where": None,
        "limit": 5,
        "result_index": [5, 6, 7, 8, 9],
        "collection": "col2",
    },
    {
        "where": "starts_with($id, 'q0')",
        "limit": None,
        "result_index": [5, 6, 7, 8, 9],
        "collection": "col2",
    },
    {
        "where": "contains($id, '2')",
        "limit": None,
        "result_index": [12],
        "collection": "col2",
        "except_providers": [
            KeyValueStoreProvider.DS_AMAZON_DYNAMODB,
            KeyValueStoreProvider.DS_AZURE_COSMOS_DB,
            KeyValueStoreProvider.DS_GOOGLE_FIRESTORE,
            KeyValueStoreProvider.DS_MONGODB,
            KeyValueStoreProvider.OS_AMAZON_S3,
            KeyValueStoreProvider.OS_AZURE_BLOB_STORAGE,
            KeyValueStoreProvider.OS_GOOGLE_CLOUD_STORAGE,
            KeyValueStoreProvider.OS_FILE_SYSTEM,
        ],
    },
]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    providers,
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
@pytest.mark.parametrize(
    "type",
    ["binary", "string"],
)
async def test_put_get_delete(provider_type: str, async_call: bool, type: str):
    client = KeyValueStoreSyncAndAsyncClient(
        provider_type=provider_type,
        async_call=async_call,
        type=type,
    )
    kv = kvs[0]
    key = get_key(kv)
    value = kv["value"]
    replace_kv = {"id": kv["id"], "value": b"replace value"}
    replace_value = replace_kv["value"]
    replace2_kv = {"id": kv["id"], "value": b"blah blah value"}
    replace2_value = replace2_kv["value"]

    await cleanup_kv(provider_type, key, client)

    # get when item doesn't exist
    with pytest.raises(NotFoundError):
        await client.get(key=key)

    # unconditional delete when item doesn't exist
    with pytest.raises(NotFoundError):
        await client.delete(key=key)

    # check exists when item doesn't exist
    response = await client.exists(key=key)
    result = response.result
    assert result is False

    # unconditional put when item doesn't exist
    response = await client.put(key=key, value=value)
    result = response.result
    assert_put_result(result, kv, provider_type)

    # get when item exists
    response = await client.get(key=key)
    result = response.result
    assert_get_result(result, kv, type, provider_type)

    # check exists when item doesn't exists
    response = await client.exists(key=key)
    result = response.result
    assert result is True

    # unconditional put when item exists
    response = await client.put(key=key, value=replace_value)
    result = response.result
    assert_put_result(result, replace_kv, provider_type)
    response = await client.get(key=key)
    result = response.result
    assert_get_result(result, replace_kv, type, provider_type)

    # range get
    response = await client.get(key=key, start=2)
    result = response.result
    assert_get_result(result, replace_kv, type, provider_type, start=2)
    response = await client.get(key=key, end=5)
    result = response.result
    assert_get_result(result, replace_kv, type, provider_type, end=5)
    response = await client.get(key=key, start=3, end=8)
    result = response.result
    assert_get_result(result, replace_kv, type, provider_type, start=3, end=8)

    # unconditional delete when item exists
    response = await client.delete(key=key)
    result = response.result
    assert_delete_result(result)

    # get when item doesn't exist
    with pytest.raises(NotFoundError):
        await client.get(key=key)

    # conditional put (exists=True) when item doesn't exist
    with pytest.raises(PreconditionFailedError):
        await client.put(key=key, value=value, where="exists()")

    # conditional put (exists=False) when item doesn't exist
    response = await client.put(key=key, value=value, where="not_exists()")
    result = response.result
    assert_put_result(result, kv, provider_type)
    response = await client.get(key=key)
    result = response.result
    assert_get_result(result, kv, type, provider_type)

    # conditional put (exists=False) when item exists
    with pytest.raises(PreconditionFailedError):
        await client.put(key=key, value=value, where="not_exists()")

    # conditional put (exists=True) when item exists
    put_response = await client.put(
        key=key, value=replace_value, where="exists()"
    )
    put_result = put_response.result
    assert_put_result(put_result, replace_kv, provider_type)

    get_response = await client.get(key=key)
    get_result = get_response.result
    assert_get_result(get_result, replace_kv, type, provider_type)

    # conditional put with bad etag
    if provider_type not in except_replace_etag_providers:
        with pytest.raises(PreconditionFailedError):
            await client.put(
                key=key,
                value=value,
                where=f"$etag='{get_bad_etag(provider_type)}'",
            )

    # conditional put with good etag
    put_response = await client.put(
        key=key,
        value=replace2_value,
        where=f"$etag='{get_result.properties.etag}'",
    )
    put_result = put_response.result
    assert_put_result(put_result, replace2_kv, provider_type)
    if provider_type not in except_get_etag_providers:
        assert get_result.properties.etag != put_result.properties.etag

    get_response = await client.get(key=key)
    get_result = get_response.result
    assert_get_result(get_result, replace2_kv, type, provider_type)

    # conditional delete with bad etag
    if provider_type not in except_delete_etag_providers:
        with pytest.raises(PreconditionFailedError):
            await client.delete(
                key=key,
                where=f"$etag='{get_bad_etag(provider_type)}'",
            )

    # conditional delete with good etag
    response = await client.delete(
        key=key, where=f"$etag='{get_result.properties.etag}'"
    )
    result = response.result
    assert_delete_result(result)

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    providers,
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_update(provider_type: str, async_call: bool):
    client = KeyValueStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call, type="binary"
    )
    for kv in kvs:
        await cleanup_kv(provider_type, get_key(kv), client)

    kv = kvs[1]
    key = kv["id"]
    value = kv["value"]

    response = await client.put(key=key, value=value)
    result = response.result
    assert result.value is None

    response = await client.get(key=key)
    result = response.result
    assert value is int(result.value)

    response = await client.put(key=key, value=30, returning="new")
    result = response.result
    assert int(result.value) == 30

    response = await client.update(key=key, set="$value=increment(5)")
    result = response.result
    assert_update_result(result, kv, provider_type)
    assert result.value is None

    response = await client.get(key=key)
    get_result = response.result
    assert int(get_result.value) == 35

    # bad etag update
    if provider_type not in except_update_etag_providers:
        with pytest.raises(PreconditionFailedError):
            response = await client.update(
                key=key,
                set="$value=increment(-10)",
                where="$etag='111'",
                returning="new",
            )

    # good etag update
    response = await client.update(
        key=key,
        set="$value=increment(-10)",
        where=f"$etag='{get_result.properties.etag}'",
        returning="new",
    )
    result = response.result
    assert int(result.value) == 25
    if provider_type not in except_get_etag_providers:
        assert result.properties.etag != get_result.properties.etag

    response = await client.get(key=key)
    result = response.result
    assert int(result.value) == 25

    if provider_type not in except_incr_float_providers:
        response = await client.update(key=key, set="$value=increment(3.5)")
        response = await client.get(key=key)
        result = response.result
        assert float(result.value) == 28.5

        response = await client.update(key=key, set="$value=increment(-20.5)")
        response = await client.get(key=key)
        result = response.result
        assert float(result.value) == 8.0

    response = await client.delete(key=key)
    result = response.result
    assert_delete_result(result)

    response = await client.update(key=key, set="$value=increment(5)")
    result = response.result
    assert_update_result(result, kv, provider_type)

    response = await client.get(key=key)
    result = response.result
    assert int(result.value) == 5

    response = await client.delete(key=key)
    result = response.result
    assert_delete_result(result)

    await client.close()

    if provider_type in except_update_string_providers:
        return

    client = KeyValueStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call, type="string"
    )

    kv = kvs[2]
    key = kv["id"]
    value = kv["value"]

    response = await client.put(key=key, value=value)
    response = await client.get(key=key)
    result = response.result
    assert result.value == value

    response = await client.update(key=key, set="$value=append('Universe')")
    response = await client.get(key=key)
    result = response.result
    assert result.value == f"{value}Universe"

    response = await client.delete(key=key)
    result = response.result
    assert_delete_result(result)

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    providers,
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_query(provider_type: str, async_call: bool):
    client = KeyValueStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call, type="binary"
    )
    for kv in query_kvs:
        try:
            await client.delete(
                key=get_key(kv),
                collection=kv["collection"],
            )
        except NotFoundError:
            pass

    for kv in query_kvs:
        await client.put(
            key=get_key(kv),
            value=get_value(kv),
            collection=kv["collection"],
        )

    for query in queries:
        if (
            "except_providers" in query
            and provider_type in query["except_providers"]
        ):
            continue
        response = await client.query(
            where=query["where"],
            limit=query["limit"],
            collection=query["collection"],
        )
        result = response.result
        if query["limit"]:
            assert len(result.items) >= len(query["result_index"])
        else:
            assert len(result.items) == len(query["result_index"])
        if not query["limit"]:
            response = await client.count(
                where=query["where"],
                collection=query["collection"],
            )
            result = response.result
            assert result == len(query["result_index"])

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    providers,
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
@pytest.mark.parametrize(
    "type",
    ["binary", "string"],
)
async def test_batch(provider_type: str, async_call: bool, type: str):
    client = KeyValueStoreSyncAndAsyncClient(
        provider_type=provider_type,
        async_call=async_call,
        type=type,
    )
    for kv in batch_kvs:
        try:
            await client.delete(
                key=get_key(kv),
                collection=kv["collection"],
            )
        except NotFoundError:
            pass

    batch = KeyValueBatch()
    for kv in batch_kvs:
        batch.put(
            key=get_key(kv),
            value=get_value(kv),
            collection=kv["collection"],
        )
    response = await client.batch(batch=batch)
    result = response.result
    for item_result, kv in zip(result, batch_kvs):
        assert_put_result(item_result, kv, provider_type)

    batch = KeyValueBatch()
    for kv in batch_kvs:
        batch.get(
            key=get_key(kv),
            collection=kv["collection"],
        )
    response = await client.batch(batch=batch)
    result = response.result
    for item_result, kv in zip(result, batch_kvs):
        assert_get_result(
            item_result,
            kv,
            type,
            provider_type,
            batch=True,
        )

    batch = KeyValueBatch()
    for kv in batch_kvs:
        batch.delete(
            key=get_key(kv),
            collection=kv["collection"],
        )
    response = await client.batch(batch=batch)
    result = response.result
    for item_result in result:
        assert_delete_result(item_result)

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    providers,
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_expire(provider_type: str, async_call: bool):
    client = KeyValueStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call, type="binary"
    )

    kv = kvs[0]
    key = get_key(kv)
    value = kv["value"]

    await cleanup_kv(provider_type, key, client)

    # get when item doesn't exist
    with pytest.raises(NotFoundError):
        await client.get(key=key)

    response = await client.put(key=key, value=value, expiry=3000)
    result = response.result
    assert_put_result(result, kv, provider_type)

    response = await client.get(key=key)
    result = response.result
    assert_get_result(result, kv, type, provider_type)

    time.sleep(4)
    # get when item expired
    with pytest.raises(NotFoundError):
        await client.get(key=key)

    # delete when item expired
    with pytest.raises(NotFoundError):
        await client.delete(key=key)

    await client.close()


def get_value(item):
    return item["value"]


def get_key(item):
    return item["id"]


def assert_get_result(
    item: KeyValueItem,
    kv: dict,
    type: str,
    provider_type: str,
    batch: bool = False,
    start: int | None = None,
    end: int | None = None,
):
    assert item.key.id == kv["id"]
    if type == "string":
        check_value = kv["value"].decode()
    else:
        check_value = kv["value"]
    if not start:
        start = 0
    if not end:
        end = len(check_value)
    else:
        end = end + 1
    check_value = check_value[start:end]
    assert item.value == check_value
    if batch and provider_type not in except_batch_get_etag_providers:
        assert item.properties.etag is not None


def assert_put_result(
    item: KeyValueItem,
    kv: dict,
    provider_type: str,
):
    assert item.key.id == kv["id"]
    if provider_type not in except_put_update_return_etag_providers:
        assert item.properties.etag is not None


def assert_update_result(
    item: KeyValueItem,
    kv: dict,
    provider_type: str,
):
    assert item.key.id == kv["id"]
    if provider_type not in except_put_update_return_etag_providers:
        assert item.properties.etag is not None


def assert_delete_result(result):
    assert result is None


def get_bad_etag(provider_type: str):
    if provider_type == KeyValueStoreProvider.MEMCACHED:
        return "111"
    return "CK/Z68ON54sDENED"


async def cleanup_kv(
    provider_type,
    key,
    client: KeyValueStoreSyncAndAsyncClient,
):
    try:
        await client.delete(key=key)
    except NotFoundError:
        return



from enum import Enum
from typing import Any, Literal, Union

from ...core import DataModel


class MatchCondition(DataModel):
    """Match condition."""

    exists: bool | None = None
    """Check item exists."""

    if_match: str | None = None
    """If match."""

    if_none_match: str | None = None
    """If none match."""

    if_modified_since: float | None = None
    """If modified since."""

    if_unmodified_since: float | None = None
    """If unmodified since."""

    if_version_match: str | None = None
    """If version match."""

    if_version_not_match: str | None = None
    """If version not match."""


class BaseIndex(DataModel):
    """Store index."""

    type: str
    """Index type."""

    name: str | None = None
    """Index name."""

    nconfig: Any | None = None
    """Native config options."""


class WildcardIndex(BaseIndex):
    """Wildcard index."""

    field: str = "*"
    """Wilcard path."""

    excluded: list[str] = []
    """Excluded paths."""

    type: Literal["wildcard"] = "wildcard"
    """Index type."""


class FieldIndex(BaseIndex):
    """Field index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["field"] = "field"
    """Index type."""


class ExcludeIndex(BaseIndex):
    """Exclude index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["exclude"] = "exclude"


class RangeIndex(BaseIndex):
    """Field index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["range"] = "range"
    """Index type."""


class HashIndex(BaseIndex):
    """Field index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["hash"] = "hash"
    """Index type."""


class AscIndex(BaseIndex):
    """Ascending index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["asc"] = "asc"
    """Index type."""


class DescIndex(BaseIndex):
    """Descending index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["desc"] = "desc"
    """Index type."""


class ArrayIndex(BaseIndex):
    """Array index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["array"] = "array"
    """Index type."""


class TTLIndex(BaseIndex):
    """TTL index."""

    field: str
    """Index field."""

    type: Literal["ttl"] = "ttl"
    """Index type."""


class GeospatialFieldType(str, Enum):
    """Geospatial field type."""

    POINT = "point"
    """Point."""

    SHAPE = "shape"
    """Shape."""


class GeospatialIndex(BaseIndex):
    """Geospatial index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["geospatial"] = "geospatial"
    """Index type."""


class TextSimilarityAlgorithm(str, Enum):
    """Text scoring algorithm for text index."""

    BM25 = "BM25"
    """BM25 algorithm."""

    BOOLEAN = "boolean"
    """Boolean similarity."""


class TextIndex(BaseIndex):
    """Text index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    variant: str | None = None
    """Text index variant."""

    similarity: TextSimilarityAlgorithm | None = None
    """Text scoring algorithm."""

    type: Literal["text"] = "text"
    """Index type."""


class VectorIndexMetric(str, Enum):
    """Vector index metric."""

    DOT_PRODUCT = "dot_product"
    COSINE = "cosine"
    EUCLIDEAN = "euclidean"
    MANHATTAN = "manhattan"
    HAMMING = "hamming"
    MAX_INNER_PRODUCT = "max_inner_product"


class VectorIndexStructure(str, Enum):
    """Vector index structure."""

    FLAT = "flat"
    HNSW = "hnsw"
    DISKANN = "diskann"

    INT8_HNSW = "int8_hnsw"
    INT4_HNSW = "int4_hnsw"
    BQ_HNSW = "bbq_hnsw"

    INT8_FLAT = "int8_flat"
    INT4_FLAT = "int4_flat"
    BQ_FLAT = "bbq_flat"
    QUANTIZED_FLAT = "quantized_flat"


class VectorIndex(BaseIndex):
    """Vector index."""

    field: str = "vector"
    """Index field name."""

    field_type: str | None = None
    """Field type."""

    dimension: int = 4
    """Vector dimension."""

    metric: VectorIndexMetric = VectorIndexMetric.DOT_PRODUCT
    """Index metric."""

    structure: VectorIndexStructure | None = None
    """Vector index structure."""

    m: int | None = None
    """The number of neighbors each node will be
    connected to in the HNSW graph."""

    ef_construction: int | None = None
    """The number of candidates to track while assembling
    the list of nearest neighbors for each new node. """

    ef_runtime: int | None = None
    """	Max top candidates during KNN search.
    Higher values increase accuracy,
    but also increase search latency."""

    epsilon: float | None = None
    """Relative factor that sets the boundaries in which a
    range query may search for candidates. That is, vector candidates
    whose distance from the query vector is radius * (1 + EPSILON) are
    potentially scanned, allowing more extensive search and more
    accurate results, at the expense of run time."""

    confidence_interval: float | None = None
    """The confidence interval to use when quantizing the vectors."""

    partitions: int | None = None
    """Partitions for IVF flat."""

    type: Literal["vector"] = "vector"
    """Index type."""


class SparseVectorIndex(BaseIndex):
    """Sparse vector index."""

    field: str = "sparse_vector"
    """Sparse vector field name."""

    field_type: str | None = None
    """Field type."""

    type: Literal["sparse_vector"] = "sparse_vector"
    """Index type."""


class CompositeIndex(BaseIndex):
    """Composite index."""

    fields: list[
        FieldIndex
        | HashIndex
        | AscIndex
        | DescIndex
        | RangeIndex
        | ArrayIndex
        | GeospatialIndex
        | TextIndex
        | VectorIndex
        | SparseVectorIndex
    ]
    """Composite index fields."""

    type: Literal["composite"] = "composite"
    """Index type."""


class RankMethod(str, Enum):
    LOG = "log"
    SATURATION = "saturation"
    LINEAR = "linear"
    FRESHNESS = "freshness"  # for Azure etc.


class RankIndex(BaseIndex):
    field: str
    method: RankMethod = RankMethod.LOG
    weight: float | None = None
    range_start: float | None = None
    range_end: float | None = None
    type: Literal["rank"] = "rank"


class IndexType:
    FIELD = "field"
    COMPOSITE = "composite"
    ASC = "asc"
    DESC = "desc"
    ARRAY = "array"
    EXCLUDE = "exclude"
    TTL = "ttl"
    GEOSPATIAL = "geospatial"
    TEXT = "text"
    VECTOR = "vector"
    SPARSE_VECTOR = "sparse_vector"
    RANGE = "range"
    HASH = "hash"
    WILDCARD = "wildcard"
    RANK = "rank"


Index = Union[
    FieldIndex,
    CompositeIndex,
    AscIndex,
    DescIndex,
    ArrayIndex,
    TTLIndex,
    GeospatialIndex,
    TextIndex,
    VectorIndex,
    SparseVectorIndex,
    RangeIndex,
    HashIndex,
    ExcludeIndex,
    WildcardIndex,
    BaseIndex,
]


class IndexStatus(str, Enum):
    """Index status."""

    CREATED = "created"
    EXISTS = "exists"
    COVERED = "covered"

    DROPPED = "dropped"
    NOT_EXISTS = "not_exists"

    NOT_SUPPORTED = "not_supported"
    ERROR = "error"


class IndexResult(DataModel):
    """Index result."""

    status: IndexStatus
    """Index status."""

    index: Index | None = None
    """Index."""

    error: str | None = None
    """Error string."""


class CollectionStatus(str, Enum):
    """Collection status"""

    CREATED = "created"
    EXISTS = "exists"

    DROPPED = "dropped"
    NOT_EXISTS = "not_exists"


class CollectionResult(DataModel):
    """Collection result."""

    status: CollectionStatus
    """Collection status."""

    indexes: list[IndexResult] = []
    """Index results."""


class FieldDefinition(DataModel):
    """Field definition."""

    attribute: str | None
    """Attribute name."""

    field: str
    """Field name."""

    field_type: str | None
    """Field type."""



from ._api_provider import APIProvider
from ._models import APIInfo, ArgMapping, ComponentMapping, OperationMapping
from .component import API

__all__ = [
    "API",
    "APIInfo",
    "APIProvider",
    "ComponentMapping",
    "OperationMapping",
    "ArgMapping",
]



"""
Docker provider for containerizer.
"""

__all__ = ["Docker"]

import os
import shutil
import subprocess
import sys
import tempfile

import docker

from x8.core import Provider, Response, RunContext
from x8.core.constants import ROOT_PACKAGE_NAME

from .._helper import create_requirements_file
from .._models import (
    BuildConfig,
    ContainerItem,
    ImageItem,
    PrepareConfig,
    RunConfig,
    SourceItem,
)


class Docker(Provider):
    def __init__(
        self,
        **kwargs,
    ):
        """Initialize."""
        super().__init__(**kwargs)

    def prepare(
        self,
        handle: str,
        config: PrepareConfig = PrepareConfig(),
        run_context: RunContext = RunContext(),
        **kwargs,
    ) -> Response[SourceItem]:
        working_dir = run_context.path
        if not config.prepare_in_place:
            temp_dir = tempfile.mkdtemp()
            shutil.copytree(run_context.path, temp_dir, dirs_exist_ok=True)
            working_dir = temp_dir

        requirements_file = f"requirements-{handle}.txt"
        create_requirements_file(
            handle,
            run_context.tag,
            requirements_file,
            working_dir,
            run_context.manifest,
        )
        requirements_files: list = [requirements_file]
        if isinstance(config.requirements, str):
            requirements_files.append(config.requirements)
        elif isinstance(config.requirements, list):
            requirements_files.extend(config.requirements)
        cmd = [ROOT_PACKAGE_NAME, "run", handle]
        if run_context.tag:
            cmd.append(run_context.tag)
        self._create_dockerfile(
            base_image=config.base_image,
            requirements_files=requirements_files,
            expose=config.expose,
            cmd=cmd,
            out_dir=working_dir,
        )
        result = SourceItem(source=working_dir)
        return Response(result=result)

    def build(
        self,
        source: str,
        config: BuildConfig = BuildConfig(),
        **kwargs,
    ) -> Response[ImageItem]:
        image_name = config.image_name or source.split("/")[-1].split(".")[0]
        if sys.platform == "darwin":
            # docker-py does not support platform parameter correctly
            # on MacOS, so we use subprocess to call docker directly
            cmd: list[str] = [
                "docker",
                "build",
                "--platform",
                config.platform,
                "--rm",
                "--load",
                "-t",
                image_name,
            ]
            if config.nocache:
                cmd += ["--no-cache"]
            cmd += [source]
            subprocess.run(
                cmd,
                check=True,
            )
            client = docker.from_env()
            image = client.images.get(image_name)
            result = self._get_image_item(image)
            return Response(result=result)
        else:
            client = docker.from_env()
            image, build_logs = client.images.build(
                path=source,
                tag=config.image_name,
                nocache=config.nocache,
                rm=True,
                platform=config.platform,
                **kwargs,
            )
            result = self._get_image_item(image)
            return Response(result=result)

    def run(
        self,
        image_name: str,
        config: RunConfig = RunConfig(),
        **kwargs,
    ) -> Response[ContainerItem]:
        client = docker.from_env()
        args = dict(
            image=image_name,
            detach=config.detach,
            remove=config.remove,
        )
        if config.ports is not None:
            args["ports"] = {
                f"{key}/tcp" if not key.endswith("/tcp") else key: value
                for key, value in config.ports.items()
            }
        if config.env is not None:
            args["environment"] = config.env
        container = client.containers.run(**args)
        result = ContainerItem(
            id=container.id,
            name=container.name,
            image=self._get_image_item(container.image),
        )
        return Response(result=result)

    def stop(
        self,
        container_id: str,
        **kwargs,
    ) -> Response[None]:
        client = docker.from_env()
        container = client.containers.get(container_id)
        container.stop()
        return Response(result=None)

    def remove(
        self,
        container_id: str,
        **kwargs,
    ) -> Response[None]:
        client = docker.from_env()
        container = client.containers.get(container_id)
        container.remove(force=True)
        return Response(result=None)

    def delete(
        self,
        image_name: str,
        digest: str | None = None,
        **kwargs,
    ) -> Response[None]:
        client = docker.from_env()
        if digest is None:
            image = client.images.get(image_name)
            client.images.remove(image.id, force=True)
        else:
            image = client.images.get(f"{image_name}@{digest}")
            client.images.remove(image.id, force=True)
        return Response(result=None)

    def tag(
        self,
        image_name: str,
        repository_name: str,
        tag: str,
        digest: str | None = None,
    ) -> Response[None]:
        client = docker.from_env()
        if digest is None:
            image = client.images.get(image_name)
            image.tag(repository_name, tag=tag)
        else:
            image = client.images.get(f"{image_name}@{digest}")
            image.tag(repository_name, tag=tag)
        return Response(result=None)

    def push(
        self,
        image_name: str,
        repository_name: str,
        digest: str | None = None,
        tag: str | None = None,
    ) -> Response[None]:
        # TODO: throw exception if there is error in push logs
        client = docker.from_env()
        if digest is None:
            image = client.images.get(image_name)
            if tag is not None:
                image.tag(repository_name, tag=tag)
            for line in client.images.push(
                repository_name, tag=tag, stream=True, decode=True
            ):

                print(line)
        else:
            image = client.images.get(f"{image_name}@{digest}")
            if tag is not None:
                image.tag(repository_name, tag=tag)
            for line in client.images.push(
                repository_name, tag=tag, stream=True, decode=True
            ):

                print(line)
        return Response(result=None)

    def pull(
        self,
        image_name: str,
        tag: str | None = None,
    ) -> Response[None]:
        client = docker.from_env()
        if tag is None:
            client.images.pull(image_name)
        else:
            client.images.pull(f"{image_name}:{tag}")
        return Response(result=None)

    def list_images(self) -> Response[list[ImageItem]]:
        client = docker.from_env()
        images = client.images.list()
        image_items = []
        for image in images:
            if image.tags:
                image_items.append(self._get_image_item(image))
        result = image_items
        return Response(result=result)

    def list_containers(self) -> Response[list[ContainerItem]]:
        client = docker.from_env()
        containers = client.containers.list()
        container_items = []
        for container in containers:
            container_items.append(
                ContainerItem(
                    id=container.id,
                    name=container.name,
                    image=self._get_image_item(
                        container.image if container.image else None,
                    ),
                )
            )
        result = container_items
        return Response(result=result)

    def _get_image_item(self, image) -> ImageItem:
        return ImageItem(
            name=image.tags[0].split(":")[0],
            digest=image.id,
            tags=image.tags,
        )

    def _create_dockerfile(
        self,
        base_image,
        requirements_files: list[str] | None,
        expose: int | list[int] | None,
        cmd: list[str] | None,
        out_dir: str,
    ):
        app_dir = "/app"
        lines = [
            f"FROM {base_image}",
            f"WORKDIR {app_dir}",
        ]
        if requirements_files is not None:
            for requirements_file in requirements_files:
                lines.append(f"COPY {requirements_file} {app_dir}")
                pip_install = "RUN pip install --no-cache-dir --upgrade"
                lines.append(f"{pip_install} -r {requirements_file}")
        lines.append(f"COPY . {app_dir}")
        expose_list = None
        if isinstance(expose, int):
            expose_list = [expose]
        elif isinstance(expose, list):
            expose_list = expose
        if expose_list:
            for port in expose_list:
                lines.append(f"EXPOSE {port}")

        if cmd is not None:
            arr = ", ".join(f'"{part}"' for part in cmd)
            lines.append(f"CMD [{arr}]")

        with open(os.path.join(out_dir, "Dockerfile"), "w") as dockerfile:
            dockerfile.write("\n\n".join(lines))

        return lines



__all__ = ["OpenAIProvider"]

from typing import Any, Mapping

from openai import AsyncOpenAI as BaseAsyncOpenAI
from openai import OpenAI as BaseOpenAI
from x8.core import Context, Provider


class OpenAIProvider(Provider):
    model: str | None
    api_key: str | None
    organization: str | None
    project: str | None
    base_url: str | None
    websocket_base_url: str | None
    webhook_secret: str | None
    timeout: float | None
    max_retries: int | None
    default_headers: Mapping[str, str] | None
    default_query: Mapping[str, object] | None
    nparams: dict | None

    _client: BaseOpenAI
    _aclient: BaseAsyncOpenAI
    _init: bool
    _ainit: bool

    def __init__(
        self,
        model: str | None,
        api_key: str | None = None,
        organization: str | None = None,
        project: str | None = None,
        base_url: str | None = None,
        websocket_base_url: str | None = None,
        webhook_secret: str | None = None,
        timeout: float | None = None,
        max_retries: int | None = None,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        nparams: dict | None = None,
        **kwargs: Any,
    ):
        """Initialize.

        Args:
            model:
                OpenAI model to use for video generation.
            api_key:
                OpenAI API key.
            organization:
                OpenAI organization.
            project:
                OpenAI project.
            base_url:
                OpenAI base url.
            websocket_base_url:
                OpenAI websocket base url.
            webhook_secret:
                OpenAI webhook secret.
            timeout:
                Timeout for client.
            max_retries:
                Maximum number of retries for failed requests.
            default_headers:
                Default headers to include in every request.
            default_query:
                Default query parameters to include in every request.
            nparams:
                Native params for OpenAI client.
        """

        self.model = model
        self.api_key = api_key
        self.organization = organization
        self.project = project
        self.base_url = base_url
        self.websocket_base_url = websocket_base_url
        self.webhook_secret = webhook_secret
        self.timeout = timeout
        self.max_retries = max_retries
        self.default_headers = default_headers
        self.default_query = default_query
        self.nparams = nparams
        self._init = False
        self._ainit = False
        super().__init__(**kwargs)

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return

        self._client = BaseOpenAI(**self._get_client_params())
        self._init = True

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._ainit:
            return

        self._aclient = BaseAsyncOpenAI(**self._get_client_params())
        self._ainit = True

    def _get_client_params(self) -> dict[str, Any]:
        params = {
            "api_key": self.api_key,
            "organization": self.organization,
            "project": self.project,
            "base_url": self.base_url,
            "websocket_base_url": getattr(self, "websocket_base_url", None),
            "webhook_secret": getattr(self, "webhook_secret", None),
            "timeout": self.timeout,
            "max_retries": self.max_retries,
            "default_headers": self.default_headers,
            "default_query": self.default_query,
        }
        return {k: v for k, v in params.items() if v is not None} | (
            self.nparams or {}
        )



from __future__ import annotations

from typing import IO, Any

from x8.core import Response, operation
from x8.ql import Expression, Value
from x8.storage._common import CollectionResult, StoreComponent

from ._models import (
    ObjectBatch,
    ObjectCollectionConfig,
    ObjectItem,
    ObjectKey,
    ObjectList,
    ObjectProperties,
    ObjectQueryConfig,
    ObjectSource,
    ObjectTransferConfig,
)


class ObjectStore(StoreComponent):
    collection: str | None

    def __init__(self, collection: str | None = None, **kwargs: Any):
        """Initialize.

        Args:
            collection:
                Collection name.
        """
        self.collection = collection
        super().__init__(**kwargs)

    @operation()
    def create_collection(
        self,
        collection: str | None = None,
        config: dict | ObjectCollectionConfig | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.
            where:
                Condition expression.

        Returns:
            Collection operation result.
        """
        raise NotImplementedError

    @operation()
    def drop_collection(
        self,
        collection: str | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Drop collection.

        Args:
            collection:
                Collection name.
            where:
                Condition expression.

        Returns:
            Collection operation result.
        """
        raise NotImplementedError

    @operation()
    def list_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        raise NotImplementedError

    @operation()
    def has_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if the collection exists.

        Args:
            collection:
                Collection name.

        Returns:
            A value indicating whether the collection exists.
        """
        raise NotImplementedError

    @operation()
    def put(
        self,
        key: str | dict | ObjectKey,
        value: Value = None,
        file: str | None = None,
        stream: IO | None = None,
        metadata: dict | None = None,
        properties: dict | ObjectProperties | None = None,
        where: str | Expression | None = None,
        config: dict | ObjectTransferConfig | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Put object.

        Args:
            key:
                Object key.
            value:
                Object value. If the type is not bytes, it
                will be converted to bytes.
            file:
                File path to upload.
            stream:
                Stream to upload.
            metadata:
                Custom metadata.
            properties:
                Object properties.
            where:
                Condition expression.
            config:
                Transfer config.
            collection:
                Collection name.

        Returns:
            Object item.

        Raises:
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    def get(
        self,
        key: str | dict | ObjectKey,
        file: str | None = None,
        stream: IO | None = None,
        where: str | Expression | None = None,
        start: int | None = None,
        end: int | None = None,
        config: dict | ObjectTransferConfig | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Get object.

        Args:
            key:
                Object key.
            file:
                File path to download.
            stream:
                Stream to download to.
            where:
                Condition expression.
            start:
                Start bytes for range request.
            end:
                End bytes for range request.
            config:
                Transfer config.
            collection:
                Collection name.

        Returns:
            Object item.

        Raises:
            NotFoundError:
                Object not found.
            NotModified:
                Object not modified.
        """
        raise NotImplementedError

    @operation()
    def get_metadata(
        self,
        key: str | dict | ObjectKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Get object metadata.

        Args:
            key:
                Object key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Object item with metadata.

        Raises:
            NotFoundError:
                Object not found.
            NotModified:
                Object not modified.
        """
        raise NotImplementedError

    @operation()
    def get_properties(
        self,
        key: str | dict | ObjectKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Get object properties.

        Args:
            key:
                Object key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Object item with properties.

        Raises:
            NotFoundError:
                Object not found.
            NotModified:
                Object not modified.
        """
        raise NotImplementedError

    @operation()
    def get_versions(
        self,
        key: str | dict | ObjectKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Get object versions.

        Args:
            key:
                Object key.
            collection:
                Collection name.

        Returns:
            Object item with object versions.

        Raises:
            NotFoundError:
                Object not found.
        """
        raise NotImplementedError

    @operation()
    def update(
        self,
        key: str | dict | ObjectKey,
        metadata: dict | None = None,
        properties: dict | ObjectProperties | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Update object metadata and properties.

        Args:
            key:
                Objecy key.
            metadata:
                Custom metadata.
            properties:
                Object properties.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Object item.

        Raises:
            NotFoundError:
                Object not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    def delete(
        self,
        key: str | dict | ObjectKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete object.

        Args:
            key:
                Object key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Object not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    def copy(
        self,
        key: str | dict | ObjectKey,
        source: str | dict | ObjectSource,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Copy object.

        Args:
            key:
                Destination object key.
            source:
                Object source.
            collection:
                Collection name.

        Returns:
            Copied object item.

        Raises:
            NotFoundError:
                Source object not found.
        """
        raise NotImplementedError

    @operation()
    def generate(
        self,
        key: str | dict | ObjectKey,
        method: str,
        expiry: int,
        collection: str | None = None,
    ) -> Response[ObjectItem]:
        """Generate signed URL.

        Args:
            key:
                Object key.
            method:
                Operation method. One of "GET", "PUT" or "DELETE".
            expiry:
                Expiry in milliseconds.
            collection:
                Collection name.

        Returns:
            Object item with signed URL.
        """
        raise NotImplementedError

    @operation()
    def query(
        self,
        where: str | Expression | None = None,
        limit: int | None = None,
        continuation: str | None = None,
        config: dict | ObjectQueryConfig | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[ObjectList]:
        """Query objects.

        Args:
            where:
                Condition expression.
            limit:
                Query limit.
            continuation:
                Continuation token.
            config:
                Query config.
            collection:
                Collection name.

        Returns:
            List of objects.
        """
        raise NotImplementedError

    @operation()
    def count(
        self,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count objects.

        Args:
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Count of objects.
        """
        raise NotImplementedError

    @operation()
    def batch(
        self,
        batch: dict | ObjectBatch,
        **kwargs,
    ) -> Response[list[Any]]:
        """Batch operation.

        Args:
            batch:
                Object batch.

        Returns:
            Batch operation results.
        """
        raise NotImplementedError

    @operation()
    def close(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close the sync client.

        Returns:
            None.
        """
        raise NotImplementedError

    @operation()
    async def acreate_collection(
        self,
        collection: str | None = None,
        config: dict | ObjectCollectionConfig | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.

        Returns:
            Collection operation result.
        """
        raise NotImplementedError

    @operation()
    async def adrop_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Drop collection.

        Args:
            collection:
                Collection name.

        Returns:
            Collection operation result.
        """
        raise NotImplementedError

    @operation()
    async def alist_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        raise NotImplementedError

    @operation()
    async def ahas_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if the collection exists.

        Args:
            collection:
                Collection name.

        Returns:
            A value indicating whether the collection exists.
        """
        raise NotImplementedError

    @operation()
    async def aput(
        self,
        key: str | dict | ObjectKey,
        value: Value = None,
        file: str | None = None,
        stream: IO | None = None,
        metadata: dict | None = None,
        properties: dict | ObjectProperties | None = None,
        where: str | Expression | None = None,
        config: dict | ObjectTransferConfig | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Put object.

        Args:
            key:
                Object key.
            value:
                Object value. If the type is not bytes, it
                will be converted to bytes.
            file:
                File path to upload.
            stream:
                Stream to upload.
            metadata:
                Custom metadata.
            properties:
                Object properties.
            where:
                Condition expression.
            config:
                Transfer config.
            collection:
                Collection name.

        Returns:
            Object item.

        Raises:
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    async def aget(
        self,
        key: str | dict | ObjectKey,
        file: str | None = None,
        stream: IO | None = None,
        where: str | Expression | None = None,
        start: int | None = None,
        end: int | None = None,
        config: dict | ObjectTransferConfig | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Get object.

        Args:
            key:
                Object key.
            file:
                File path to download.
            stream:
                Stream to download to.
            where:
                Condition expression.
            start:
                Start bytes for range request.
            end:
                End bytes for range request.
            config:
                Transfer config.
            collection:
                Collection name.

        Returns:
            Object item.

        Raises:
            NotFoundError:
                Object not found.
            NotModified:
                Object not modified.
        """
        raise NotImplementedError

    @operation()
    async def aget_metadata(
        self,
        key: str | dict | ObjectKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Get object metadata.

        Args:
            key:
                Object key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Object item with metadata.

        Raises:
            NotFoundError:
                Object not found.
            NotModified:
                Object not modified.
        """
        raise NotImplementedError

    @operation()
    async def aget_properties(
        self,
        key: str | dict | ObjectKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Get object properties.

        Args:
            key:
                Object key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Object item with properties.

        Raises:
            NotFoundError:
                Object not found.
            NotModified:
                Object not modified.
        """
        raise NotImplementedError

    @operation()
    async def aget_versions(
        self,
        key: str | dict | ObjectKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Get object versions.

        Args:
            key:
                Object key.
            collection:
                Collection name.

        Returns:
            Object item with object versions.

        Raises:
            NotFoundError:
                Object not found.
        """
        raise NotImplementedError

    @operation()
    async def aupdate(
        self,
        key: str | dict | ObjectKey,
        metadata: dict | None = None,
        properties: dict | ObjectProperties | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Update object metadata and properties.

        Args:
            key:
                Objecy key.
            metadata:
                Custom metadata.
            properties:
                Object properties.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Object item.

        Raises:
            NotFoundError:
                Object not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    async def adelete(
        self,
        key: str | dict | ObjectKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete object.

        Args:
            key:
                Object key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Object not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    async def acopy(
        self,
        key: str | dict | ObjectKey,
        source: str | dict | ObjectSource,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[ObjectItem]:
        """Copy object.

        Args:
            key:
                Destination object key.
            source:
                Object source.
            collection:
                Collection name.

        Returns:
            Copied object item.

        Raises:
            NotFoundError:
                Source object not found.
        """
        raise NotImplementedError

    @operation()
    async def agenerate(
        self,
        key: str | dict | ObjectKey,
        method: str,
        expiry: int,
        collection: str | None = None,
    ) -> Response[ObjectItem]:
        """Generate signed URL.

        Args:
            key:
                Object key.
            method:
                Operation method. One of "GET", "PUT" or "DELETE".
            expiry:
                Expiry in milliseconds.
            collection:
                Collection name.

        Returns:
            Object item with signed URL.
        """
        raise NotImplementedError

    @operation()
    async def aquery(
        self,
        where: str | Expression | None = None,
        limit: int | None = None,
        continuation: str | None = None,
        config: dict | ObjectQueryConfig | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[ObjectList]:
        """Query objects.

        Args:
            where:
                Condition expression.
            limit:
                Query limit.
            continuation:
                Continuation token.
            config:
                Query config.
            collection:
                Collection name.

        Returns:
            List of objects.
        """
        raise NotImplementedError

    @operation()
    async def acount(
        self,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count objects.

        Args:
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Count of objects.
        """
        raise NotImplementedError

    @operation()
    async def abatch(
        self,
        batch: dict | ObjectBatch,
        **kwargs,
    ) -> Response[list[Any]]:
        """Batch operation.

        Args:
            operations:
                Object batch.

        Returns:
            Batch operation results.
        """
        raise NotImplementedError

    @operation()
    async def aclose(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close the async client.

        Returns:
            None.
        """
        raise NotImplementedError



from typing import Any

from x8.core import Component, Response, operation
from x8.messaging._common import (
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    QueueConfig,
    QueueInfo,
)
from x8.ql import Expression


class Queue(Component):
    queue: str | None

    def __init__(
        self,
        queue: str | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            queue:
                Queue name.
        """
        self.queue = queue
        super().__init__(**kwargs)

    @operation()
    def create_queue(
        self,
        queue: str | None = None,
        config: dict | QueueConfig | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create queue.

        Args:
            queue:
                Queue name.
            config:
                Queue config.
            where:
                Condition expression.
        """
        raise NotImplementedError

    @operation()
    def update_queue(
        self,
        queue: str | None = None,
        config: dict | QueueConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Update queue.

        Args:
            queue:
                Queue name.
            config:
                Queue config.
        """
        raise NotImplementedError

    @operation()
    def drop_queue(
        self,
        queue: str | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Drop queue.

        Args:
            queue:
                Queue name.
            where:
                Condition expression.
        """
        raise NotImplementedError

    @operation()
    def list_queues(self, **kwargs: Any) -> Response[list[str]]:
        """List queues.

        Returns:
            List of queue names.
        """
        raise NotImplementedError

    @operation()
    def has_queue(
        self,
        queue: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if queue exists.

        Args:
            queue:
                Queue name.
        Returns:
            True if queue exists, False otherwise.
        """
        raise NotImplementedError

    @operation()
    def get_queue(
        self,
        queue: str | None = None,
        **kwargs: Any,
    ) -> Response[QueueInfo]:
        """Get queue info.

        Args:
            queue:
                Queue name.

        Returns:
            Queue info.
        """
        raise NotImplementedError

    @operation()
    def put(
        self,
        value: MessageValueType,
        metadata: dict | None = None,
        properties: dict | MessageProperties | None = None,
        config: dict | MessagePutConfig | None = None,
        queue: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Put message in the queue.

        Args:
            value:
                Message value.
            metadata:
                Message metadata.
            properties:
                Message properties.
            config:
                Put config.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    def batch(
        self,
        batch: dict | MessageBatch,
        queue: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Put batch of message in the queue.

        Args:
            batch: Message batch.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    def pull(
        self,
        config: dict | MessagePullConfig | None = None,
        queue: str | None = None,
    ) -> Response[list[MessageItem]]:
        """Pull messages from the queue.

        Args:
            config:
                Pull config. Defaults to None.
            queue:
                Queue name.

        Returns:
            Pulled message items.
        """
        raise NotImplementedError

    @operation()
    def ack(
        self,
        key: dict | MessageKey,
        queue: str | None = None,
    ) -> Response[None]:
        """Acknowledge message.

        Args:
            key:
                Message key.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    def nack(
        self,
        key: dict | MessageKey,
        queue: str | None = None,
    ) -> Response[None]:
        """Abandon message.

        Args:
            key:
                Message key.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    def extend(
        self,
        key: dict | MessageKey,
        timeout: int | None = None,
        queue: str | None = None,
    ) -> Response[MessageItem]:
        """Extend message timeout.

        Args:
            key:
                Message key.
            timeout:
                Number of seconds to extend the message timeout.
            queue:
                Queue name.

        Returns:
            Message item.
        """
        raise NotImplementedError

    @operation()
    def purge(
        self,
        config: dict | MessagePullConfig | None = None,
        queue: str | None = None,
    ) -> Response[None]:
        """Purge all messages in the queue.

        Args:
            config:
                Pull config.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    def close(
        self,
        queue: str | None = None,
    ) -> Response[None]:
        """Close the queue and the client.

        Args:
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    async def acreate_queue(
        self,
        queue: str | None = None,
        config: dict | QueueConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create queue.

        Args:
            queue:
                Queue name.
            config:
                Queue config.
        """
        raise NotImplementedError

    @operation()
    async def aupdate_queue(
        self,
        queue: str | None = None,
        config: dict | QueueConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Update queue.

        Args:
            queue:
                Queue name.
            config:
                Queue config.
        """
        raise NotImplementedError

    @operation()
    async def adrop_queue(
        self, queue: str | None = None, **kwargs: Any
    ) -> Response[None]:
        """Drop queue.

        Args:
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    async def alist_queues(self, **kwargs: Any) -> Response[list[str]]:
        """List queues.

        Returns:
            List of queue names.
        """
        raise NotImplementedError

    @operation()
    async def ahas_queue(
        self,
        queue: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if queue exists.

        Args:
            queue:
                Queue name.
        Returns:
            True if queue exists, False otherwise.
        """
        raise NotImplementedError

    @operation()
    async def aget_queue(
        self,
        queue: str | None = None,
        **kwargs: Any,
    ) -> Response[QueueInfo]:
        """Get queue info.

        Args:
            queue:
                Queue name.

        Returns:
            Queue info.
        """
        raise NotImplementedError

    @operation()
    async def aput(
        self,
        value: MessageValueType,
        metadata: dict | None = None,
        properties: dict | MessageProperties | None = None,
        config: dict | MessagePutConfig | None = None,
        queue: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Put message in the queue.

        Args:
            value:
                Message value.
            metadata:
                Message metadata.
            properties:
                Message properties.
            config:
                Put config.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    async def abatch(
        self,
        batch: dict | MessageBatch,
        queue: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Put batch of message in the queue.

        Args:
            batch: Message batch.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    async def apull(
        self,
        config: dict | MessagePullConfig | None = None,
        queue: str | None = None,
    ) -> Response[list[MessageItem]]:
        """Pull messages from the queue.

        Args:
            config:
                Pull config. Defaults to None.
            queue:
                Queue name.

        Returns:
            Pulled message items.
        """
        raise NotImplementedError

    @operation()
    async def aack(
        self,
        key: dict | MessageKey,
        queue: str | None = None,
    ) -> Response[None]:
        """Acknowledge message.

        Args:
            key:
                Message key.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    async def anack(
        self,
        key: dict | MessageKey,
        queue: str | None = None,
    ) -> Response[None]:
        """Abandon message.

        Args:
            key:
                Message key.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    async def aextend(
        self,
        key: dict | MessageKey,
        timeout: int | None = None,
        queue: str | None = None,
    ) -> Response[MessageItem]:
        """Extend message timeout.

        Args:
            key:
                Message key.
            timeout:
                Number of seconds to extend the message timeout.
            queue:
                Queue name.

        Returns:
            Message item.
        """
        raise NotImplementedError

    @operation()
    async def apurge(
        self,
        config: dict | MessagePullConfig | None = None,
        queue: str | None = None,
    ) -> Response[None]:
        """Purge all messages in the queue.

        Args:
            config:
                Pull config.
            queue:
                Queue name.
        """
        raise NotImplementedError

    @operation()
    async def aclose(
        self,
        queue: str | None = None,
    ) -> Response[None]:
        """Close the queue and the client.

        Args:
            queue:
                Queue name.
        """
        raise NotImplementedError



from typing import Any

from x8.core import Response
from x8.storage.object_store import ObjectItem, ObjectList

from ._providers import get_component


class ObjectStoreSyncAndAsyncClient:
    def __init__(self, provider_type: str, async_call: bool):
        self.client = get_component(provider_type)
        self.async_call = async_call
        self.provider_type = provider_type

    async def create_collection(
        self, collection=None, config=None, **kwargs
    ) -> Response[None]:
        if self.async_call:
            return await self.client.acreate_collection(
                collection=collection, config=config, **kwargs
            )
        return self.client.create_collection(
            collection=collection, config=config, **kwargs
        )

    async def drop_collection(
        self, collection=None, **kwargs
    ) -> Response[None]:
        if self.async_call:
            return await self.client.adrop_collection(
                collection=collection, **kwargs
            )
        return self.client.drop_collection(collection=collection, **kwargs)

    async def list_collections(self, **kwargs) -> Response[list[str]]:
        if self.async_call:
            return await self.client.alist_collections(**kwargs)
        return self.client.list_collections(**kwargs)

    async def has_collection(
        self, collection=None, **kwargs
    ) -> Response[bool]:
        if self.async_call:
            return await self.client.ahas_collection(
                collection=collection, **kwargs
            )
        return self.client.has_collection(collection=collection, **kwargs)

    async def put(
        self,
        key,
        value=None,
        file=None,
        stream=None,
        metadata=None,
        properties=None,
        where=None,
        collection=None,
        config=None,
        **kwargs,
    ) -> Response[ObjectItem]:
        if self.async_call:
            return await self.client.aput(
                key=key,
                value=value,
                file=file,
                stream=stream,
                metadata=metadata,
                properties=properties,
                where=where,
                collection=collection,
                config=config,
                **kwargs,
            )
        return self.client.put(
            key=key,
            value=value,
            file=file,
            stream=stream,
            metadata=metadata,
            properties=properties,
            where=where,
            collection=collection,
            config=config,
            **kwargs,
        )

    async def get(
        self,
        key,
        file=None,
        stream=None,
        where=None,
        start=None,
        end=None,
        config=None,
        collection=None,
        **kwargs,
    ) -> Response[ObjectItem]:
        if self.async_call:
            return await self.client.aget(
                key=key,
                file=file,
                stream=stream,
                where=where,
                start=start,
                end=end,
                config=config,
                collection=collection,
                **kwargs,
            )
        return self.client.get(
            key=key,
            file=file,
            stream=stream,
            where=where,
            start=start,
            end=end,
            config=config,
            collection=collection,
            **kwargs,
        )

    async def get_metadata(
        self,
        key,
        where=None,
        collection=None,
        **kwargs,
    ) -> Response[ObjectItem]:
        if self.async_call:
            return await self.client.aget_metadata(
                key=key, where=where, collection=collection, **kwargs
            )
        return self.client.get_metadata(
            key=key, where=where, collection=collection, **kwargs
        )

    async def get_properties(
        self,
        key,
        where=None,
        collection=None,
        **kwargs,
    ) -> Response[ObjectItem]:
        if self.async_call:
            return await self.client.aget_properties(
                key=key, where=where, collection=collection, **kwargs
            )
        return self.client.get_properties(
            key=key, where=where, collection=collection, **kwargs
        )

    async def get_versions(
        self, key, collection=None, **kwargs
    ) -> Response[ObjectItem]:
        if self.async_call:
            return await self.client.aget_versions(
                key=key, collection=collection, **kwargs
            )
        return self.client.get_versions(
            key=key, collection=collection, **kwargs
        )

    async def update(
        self,
        key,
        metadata=None,
        properties=None,
        where=None,
        collection=None,
        **kwargs,
    ) -> Response[ObjectItem]:
        if self.async_call:
            return await self.client.aupdate(
                key=key,
                metadata=metadata,
                properties=properties,
                where=where,
                collection=collection,
                **kwargs,
            )
        return self.client.update(
            key=key,
            metadata=metadata,
            properties=properties,
            where=where,
            collection=collection,
            **kwargs,
        )

    async def delete(
        self, key=None, where=None, collection=None, **kwargs
    ) -> Response[None]:
        if self.async_call:
            return await self.client.adelete(
                key=key, where=where, collection=collection, **kwargs
            )
        return self.client.delete(
            key=key, where=where, collection=collection, **kwargs
        )

    async def copy(
        self, key, source, where=None, collection=None, **kwargs
    ) -> Response[ObjectItem]:
        if self.async_call:
            return await self.client.acopy(
                key=key,
                source=source,
                where=where,
                collection=collection,
                **kwargs,
            )
        else:
            return self.client.copy(
                key=key,
                source=source,
                where=where,
                collection=collection,
                **kwargs,
            )

    async def generate(
        self,
        key,
        method,
        expiry,
        collection=None,
    ) -> Response[ObjectItem]:
        if self.async_call:
            return await self.client.agenerate(
                key=key,
                method=method,
                expiry=expiry,
                collection=collection,
            )
        else:
            return self.client.generate(
                key=key,
                method=method,
                expiry=expiry,
                collection=collection,
            )

    async def query(
        self,
        where=None,
        limit=None,
        continuation=None,
        config=None,
        collection=None,
        **kwargs,
    ) -> Response[ObjectList]:
        if self.async_call:
            return await self.client.aquery(
                where=where,
                limit=limit,
                continuation=continuation,
                config=config,
                collection=collection,
            )
        else:
            return self.client.query(
                where=where,
                limit=limit,
                continuation=continuation,
                config=config,
                collection=collection,
            )

    async def count(
        self,
        where=None,
        collection=None,
        **kwargs,
    ) -> Response[int]:
        if self.async_call:
            return await self.client.acount(
                where=where,
                collection=collection,
            )
        else:
            return self.client.count(
                where=where,
                collection=collection,
            )

    async def batch(self, batch, **kwargs) -> Response[list[Any]]:
        if self.async_call:
            return await self.client.abatch(batch=batch, **kwargs)
        return self.client.batch(batch=batch, **kwargs)

    async def close(self) -> Response[None]:
        if self.async_call:
            return await self.client.aclose()
        else:
            return self.client.close()



from x8.core import DataModel


class LiveStreamParam(DataModel):
    url: str | None = None
    """Live stream url."""

    min_resolution: int = 0
    """Minimum resolution of the stream to decode."""


class LiveStreamInfo(DataModel):
    width: int
    """Width of the video in pixels."""

    height: int
    """Height of the video in pixels."""

    total_frames: int
    """Total number of frames in the video stream."""

    frame_rate: float | None = None
    """Frame rate of the video."""

    format: str | None = None
    """Format of the video."""

    codec: str | None = None
    """Name of the codec used for the video stream."""

    pixel_format: str | None = None
    """Pixel format of the video stream."""

    start_time: float | None = None
    """Start time of the video in seconds."""



import inspect
from typing import Any


class SyncAndAsyncClient:
    client: Any
    async_call: bool

    async def _execute_method(self, **kwargs):
        caller_frame = inspect.stack()[1]
        method_name = caller_frame.function
        client_method_name = (
            f"a{method_name}" if self.async_call else method_name
        )
        if method_name.startswith("__") and method_name.endswith("__"):
            method_name = method_name[2:-2]
            client_method_name = (
                f"__a{method_name}__"
                if self.async_call
                else f"__{method_name}__"
            )
        method = getattr(
            self.client,
            client_method_name,
        )
        if self.async_call:
            return await method(**kwargs)
        return method(**kwargs)
